"""
Defaults module.

Provides constants, dictionaries and default values
to be used in the project.

Since:
  2024/02/06

Authors:
  - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""

THOUSAND_SEPARATOR = ' '
UNIT_SEPARATOR = ' '
_default_interface = None

multiplier_symbols = {
    "micro": {
        "preferred": "u",
        "possibles": ['u', 'µ', 'μ']
    },
}
"""
dict: A dictionary of symbols that can be used to replace possible multiplier symbols e.g. greek characters
with preferred characters.
"""

def set_default_interface(interface):
    """Sets the default backend interface.
    
    Args:
        interface (BackendInterface): The interface to be set as the default.
    """
    global _default_interface
    _default_interface = interface

def get_default_interface():
    """Returns the default backend interface.

    Returns:
        BackendInterface: The default backend interface.
    """
    global _default_interface
    return _default_interface


"""Tests module for scientific-spinbox.

This module aggregates tests for each part of scientific-spinbox,
like backend, validators and widget. Uses pytest as the test framework.

By running tests, we ensure that all core features and behaviour are
working as expected.

Authors:
    - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""


from qtpy import QtWidgets
from decimal import Decimal, setcontext

from scientific_spinbox.widget import ScientificSpinBox
from test_gui.ui.dlg_testspinbox_externals import Ui_Dialog as Ui_TestSpinBoxExternalsDialog

class TestExternalsDialog(QtWidgets.QDialog):
    """
    Dialog that shows ScientificSpinBox and some external test controls.

    Used to test the behaviour of the ScientificSpinBox.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui = Ui_TestSpinBoxExternalsDialog()

        # Configures Ui
        self.ui.setupUi(self)

        self._controls = {
            'thousand_separators': True,
            'default_decimals': 2,
            'default_value': 0.000,
            'display_unit': 'ms',
            'enable_tooltip': True,
            'base_unit': 'us',
            'allowed_units': None,
            'preferred_units': None
        }

        self.scientificSpinBox = ScientificSpinBox(
            **self._controls,
            parent=self.ui.testSpinBox1.parentWidget()
        )

        # Replaces necessary widgets
        self.ui.testSpinBox1.parentWidget().layout().replaceWidget(
            self.ui.testSpinBox1,
            self.scientificSpinBox
        )
        self.ui.testSpinBox1 = self.scientificSpinBox
        setcontext(self.scientificSpinBox.backend.decimalsContext)

        # Connect apply button
        self.ui.applyButton.clicked.connect(self.applyControls)

        # Connect externals buttons
        self.ui.setValueFloatButton.clicked.connect(self.onSetValueFloatSlot)
        self.ui.setValueDecimalButton.clicked.connect(self.onSetValueDecimalSlot)
        self.ui.setDecimalsButton.clicked.connect(self.onSetDecimalsChangeSlot)
        self.ui.setDisplayUnitButton.clicked.connect(self.onSetDisplayUnitSlot)

        # Set defaults
        self.ui.displayUnitTextEdit.setText(self._controls['display_unit'])
        self.ui.baseUnitTextEdit.setText(self._controls['base_unit'])
        self.ui.defaultDecimalsTextEdit.setText(str(self._controls['default_decimals']))
        self.ui.defaultValueTextEdit.setText(str(self._controls['default_value']))
        self.ui.preferredUnitsTextEdit.setText(self._controls['preferred_units'])

        self.ui.setValueFloatTextEdit.setText(str(self._controls['default_value']))
        self.ui.setValueDecimalTextEdit.setText(str(self._controls['default_value']))
        self.ui.setDecimalsTextEdit.setText(str(self._controls['default_decimals']))
        self.ui.setDisplayUnitTextEdit.setText(self._controls['display_unit'])

        # Connect controls and labels
        self.connectSlots()
        self.onBaseUnitChangeSlot()
        self.onValueChangeSlot()
        self.onBaseQuantityChangeSlot()
        self.onBaseValueChangeSlot()
        self.onDisplayUnitChangeSlot()
        self.onTextChangeSlot()
        return

    def connectSlots(self):
        # Text edits
        self.ui.displayUnitTextEdit.textChanged.connect(self.onDisplayUnitChangeSlot)
        self.ui.baseUnitTextEdit.textChanged.connect(self.onBaseUnitChangeSlot)
        self.ui.defaultDecimalsTextEdit.textChanged.connect(self.onDefaultDecimalsChangeSlot)
        self.ui.defaultValueTextEdit.textChanged.connect(self.onDefaultValueChangeSlot)
        self.ui.allowedUnitsTextEdit.textChanged.connect(self.onAllowedUnitsChangeSlot)
        self.ui.preferredUnitsTextEdit.textChanged.connect(self.onPreferredUnitsChangeSlot)

        # Check boxes
        self.ui.thousandSepCheckBox.stateChanged.connect(self.onThousandSepEnableChangeSlot)
        self.ui.tooltipCheckBox.stateChanged.connect(self.onTooltipEnableChangeSlot)
        self.ui.integerRoundCheckBox.stateChanged.connect(self.onIntegerRoundCheckBoxChangeSlot)

        self.scientificSpinBox.valueChanged.connect(self.onValueChangeSlot)
        self.scientificSpinBox.textChanged.connect(self.onTextChangeSlot)
        self.scientificSpinBox.baseQuantityChanged.connect(self.onBaseQuantityChangeSlot)
        self.scientificSpinBox.baseValueChanged.connect(self.onBaseValueChangeSlot)
        self.scientificSpinBox.displayUnitChanged.connect(self.onSpinboxDisplayUnitChange)
        self.onIntegerRoundCheckBoxChangeSlot()
        return
    
    def onSetDisplayUnitSlot(self):
        text = self.ui.setDisplayUnitTextEdit.toPlainText()
        if text:
            try:
                unit = text
                self.scientificSpinBox.setDisplayUnit(unit)
            except Exception:
                return
    
    def onSetValueFloatSlot(self):
        text = self.ui.setValueFloatTextEdit.toPlainText()
        if text:
            try:
                value = float(text)
                self.scientificSpinBox.setValue(value)
            except Exception:
                return
            
    def onSetValueDecimalSlot(self):
        text = self.ui.setValueDecimalTextEdit.toPlainText()
        if text:
            try:
                value = Decimal(text)
                self.scientificSpinBox.setValue(value)
            except Exception:
                return
            
    def onSetDecimalsChangeSlot(self):
        text = self.ui.setDecimalsTextEdit.toPlainText()
        if text:
            try:
                decimals = int(text)
                self.scientificSpinBox.setDecimals(decimals)
            except Exception:
                return
    
    def onPreferredUnitsChangeSlot(self):
        if self.ui.preferredUnitsTextEdit.toPlainText():
            self._controls['preferred_units'] = self.ui.preferredUnitsTextEdit.toPlainText().split(',')
    
    def onIntegerRoundCheckBoxChangeSlot(self):
        if self.ui.integerRoundCheckBox.isChecked():
            self._controls['allowed_units'] = None
            self._controls['display_unit'] = None

            # Disable displayUnit
            self.ui.displayUnitTextEdit.setText('')
            self.ui.displayUnitTextEdit.setDisabled(True)

            # Disable allowedUnits
            self.ui.allowedUnitsTextEdit.setText('')
            self.ui.allowedUnitsTextEdit.setDisabled(True)

            # Enable preferredUnits
            self.ui.preferredUnitsTextEdit.setDisabled(False)
        else:
            # Disable preferredUnits
            self._controls['preferred_units'] = None
            self.ui.preferredUnitsTextEdit.setDisabled(True)

            # Enable allowedUnits and displayUnits
            self.ui.allowedUnitsTextEdit.setDisabled(False)
            self.ui.displayUnitTextEdit.setDisabled(False)

    
    def onDefaultValueChangeSlot(self):
        if self.ui.defaultValueTextEdit.toPlainText():
            self._controls['default_value'] = float(self.ui.defaultValueTextEdit.toPlainText())

    def onDefaultDecimalsChangeSlot(self):
        if self.ui.defaultDecimalsTextEdit.toPlainText():
            self._controls['default_decimals'] = int(self.ui.defaultDecimalsTextEdit.toPlainText())
    
    def onAllowedUnitsChangeSlot(self):
        if self.ui.allowedUnitsTextEdit.toPlainText():
            self._controls['allowed_units'] = self.ui.allowedUnitsTextEdit.toPlainText().split(',')

    def onTooltipEnableChangeSlot(self):
        self._controls['enable_tooltip'] = self.ui.tooltipCheckBox.isChecked()

    def onThousandSepEnableChangeSlot(self):
        self._controls['thousand_separators'] = self.ui.thousandSepCheckBox.isChecked()

    def onDisplayUnitChangeSlot(self):
        if self.ui.displayUnitTextEdit.toPlainText():
            self._controls['display_unit'] = self.ui.displayUnitTextEdit.toPlainText()
        else:
            self._controls['display_unit'] = None
        
        self.ui.displayUnitChangedLabel.setText("displayUnitChanged: " + str(self.scientificSpinBox.unit))

    def onBaseUnitChangeSlot(self):
        self._controls['base_unit'] = self.ui.baseUnitTextEdit.toPlainText()
        self.ui.baseUnitLabel.setText("baseUnit: " + str(self.scientificSpinBox.baseUnit))
        
    def applyControls(self):
        print(self._controls)
        self.scientificSpinBox = ScientificSpinBox(
            **self._controls,
            parent=self.ui.testSpinBox1.parentWidget()
        )

        # Replaces necessary widgets
        self.ui.testSpinBox1.parentWidget().layout().replaceWidget(
            self.ui.testSpinBox1,
            self.scientificSpinBox
        )
        self.ui.testSpinBox1 = self.scientificSpinBox
        self.connectSlots()

    def onValueChangeSlot(self):
        self.ui.valueChangedLabel.setText("valueChanged: " + str(self.scientificSpinBox.value))
    
    def onBaseValueChangeSlot(self):
        self.ui.baseValueLabel.setText("baseValue: " + str(self.scientificSpinBox.baseValue))
        self.ui.baseValueChangedLabel.setText("baseValueChanged: " + str(self.scientificSpinBox.baseValue))
        
    def onBaseQuantityChangeSlot(self):
        self.ui.baseQuantityLabel.setText("baseQuantity: " + str(self.scientificSpinBox.baseQuantity))

    def onTextChangeSlot(self):
        self.ui.textChangedLabel.setText("textChanged: " + str(self.scientificSpinBox.text()))
        
    def onSpinboxDisplayUnitChange(self):
        self.ui.displayUnitChangedLabel.setText("displayUnitChanged: " + str(self.scientificSpinBox.unit))


from typing import List
from decimal import Decimal, ROUND_UP, setcontext
from functools import partial

from scientific_spinbox.backend.interfaces import PintInterface
from scientific_spinbox.validators import ScientificInputValidator

def quantityChangeUnitFormatter(backend: PintInterface,
                                input_validator: ScientificInputValidator,
                                default_decimals: int,
                                unit_separator: str,
                                old_quantity: PintInterface.QuantityType,
                                new_quantity: PintInterface.QuantityType) -> PintInterface.QuantityType:
    """
    Formats `new_quantity` to match the decimal precision of `old_quantity`.

    Args:
        backend (PintInterface): Interface for unit and quantity operations.
        input_validator (ScientificInputValidator): Validator for parsing numeric values.
        default_decimals (int): Default decimal places if `old_quantity` has none.
        unit_separator (str): Text separator for units.
        old_quantity (PintInterface.QuantityType): Quantity to determine initial precision.
        new_quantity (PintInterface.QuantityType): Quantity to reformat.

    Returns:
        PintInterface.QuantityType: Formatted `new_quantity` with matched decimal precision.
    """
    setcontext(backend.decimalsContext)

    # Gets decimal places of original quantity
    old_quantity_text = backend.quantityTextRepr(old_quantity, unit_separator)
    old_captured = input_validator.getCapturedGroups(old_quantity_text)
    old_real_text = f"{Decimal(old_captured.scientificRepresentation):f}"
    old_decimals = len(input_validator.getCapturedGroups(old_real_text).decimalNumbers)

    # Creates a Decimal only to work with decimal places
    old_unit = backend.getQuantityUnitStr(old_quantity)
    new_unit = backend.getQuantityUnitStr(new_quantity)
    if old_decimals is None or old_decimals == 0:
        old_decimals = default_decimals

    control_places = Decimal("0." + '0' * (old_decimals - 1) + '1')
    control_quantity = backend.quantityFromDecimal(control_places, old_unit)
    new_control_quantity = backend.changeQuantityUnit(control_quantity, new_unit)
    new_control_places = Decimal(f"{backend.getQuantityValueNumeric(new_control_quantity):f}")
    captured_control = input_validator.getCapturedGroups(f"{new_control_places:f}")
    new_decimals = len(captured_control.decimalNumbers)
    #_logger.debug(f"{captured_control=}, {new_decimals=}, {new_control_places=}")

    if new_decimals is not None and new_decimals == 0:
        new_decimals = default_decimals

    if new_decimals:
        # Formats the new_quantity
        new_quantity_text = backend.quantityTextRepr(new_quantity, unit_separator)
        new_captured = input_validator.getCapturedGroups(new_quantity_text)
        new_scientific = new_captured.scientificRepresentation
        formatted_scientific = f"{Decimal(new_scientific).quantize(new_control_places, ROUND_UP):f}"
        formatted_quantity = backend.quantityFromText(formatted_scientific, new_unit)
        #_logger.debug(f"{new_decimals=}, {old_quantity_text=}, {new_control_quantity=}, {new_control_places=}, {new_scientific=}, {formatted_quantity=}")
        return formatted_quantity
    else:
        return new_quantity

def convert_to_nearest_preferred_unit(
        quantity: PintInterface.QuantityType,
        preferred_units: List[str],
        unit_separator: str,
        default_decimals: int,
        validator: ScientificInputValidator,
        interface: PintInterface
    ) -> PintInterface.QuantityType:
    """
    Converts a PintInterface's quantity to the nearest preferred unit.

    Args:
        quantity (PintInterface.QuantityType): quantity object.
        preferred_units (list[str]): list of preferred units.
        unit_separator (str): unit separator.
        validator (ScientificInputValidator): validator object.
        interface (PintInterface): interface object.

    Returns:
        PintInterface.QuantityType: converted quantity object.
    """
    selected_quantity = quantity
    
    # Formatter setup
    mockedQuantityChangeUnitFormatter = partial(
        quantityChangeUnitFormatter,
        interface,
        validator,
        default_decimals,
        unit_separator
    )

    # Populate the data list with converted quantities and their integer/decimal sizes
    data = []
    for unit in preferred_units:
        converted_quantity = interface.changeQuantityUnit(
            quantity, unit, formatter=mockedQuantityChangeUnitFormatter
        )
        captured = validator.getCapturedGroups(
            interface.quantityTextRepr(
                converted_quantity,
                unit_separator,
                normalize=True,
                formatter=lambda x: f"{x:f}"
            )
        )
        integer_part = captured.integerNumbers
        decimal_part = captured.decimalNumbers
        data.append({
            'integer_size': len(integer_part) if integer_part and int(integer_part) != 0 else 0,
            'decimal_size': len(decimal_part) if decimal_part and int(decimal_part) != 0 else 0,
            'converted_quantity': converted_quantity
        })

    # Filter for entries with non-zero integer size and sort by integer size
    nonzero_sorted_int = sorted(
        (item for item in data if item['integer_size'] > 0),
        key=lambda x: x['integer_size']
    )

    if nonzero_sorted_int:
        # Use the smallest integer size if available
        selected_quantity = nonzero_sorted_int[0]['converted_quantity']
    else:
        # If all integer sizes are zero, sort by decimal size
        nonzero_sorted_decimal = sorted(
            (item for item in data if item['decimal_size'] > 0),
            key=lambda x: x['decimal_size']
        )
        selected_quantity = (
            nonzero_sorted_decimal[0]['converted_quantity']
            if nonzero_sorted_decimal else quantity
        )

    # Return the quantity in the selected unit
    return interface.changeQuantityUnit(quantity, interface.getQuantityUnitStr(selected_quantity))



from abc import ABC, abstractmethod
from typing import Union, List
from decimal import Decimal, Context, setcontext

import pint

from ..errors.backend import (
    IncompatibleConversionUnitError,
    EmptyArrayError,
    InvalidUnitError
)

from ..defaults import multiplier_symbols

class BackendInterface(ABC):
    """
    Abstract interface between the unit-handling backend and the ScientificSpinbox

    Requires the implementation of methods for handling quantities, units and their
    representations in numeric and text forms.
    """
    @abstractmethod
    def __init__(self, unit_system: str, precision: int):
        """
        Args:
            unit_system (str): unit system to be used (SI, kgs, ...).
            precision (int): precision to be used in the numeric representation.
        """
        pass
    
    @property
    @abstractmethod
    def unitSystem() -> str:
        """str: Unit System to be used."""
        pass

    @property
    @abstractmethod
    def unitRegistry():
        """object: Unit Registry object."""
        pass
    
    @property
    @abstractmethod
    def precision() -> int:
        """int: precision of the numeric representation."""
        pass

    @abstractmethod
    def quantityFromText(self, text: str):
        """
        Creates a new quantity object based on the text representation.

        Args:
            text (str): text representation of the new quantity.

        Returns:
            object: quantity object created from text.
        """
        pass

    @abstractmethod
    def isUnitRegistered(self, unit) -> bool:
        """Verifies if a unit is registered in the Unit Registry.
        
        Args:
            unit (str): unit representation in text.

        Returns:
            bool: True if it is registered, False otherwise.
        """
        pass
    
    @abstractmethod
    def unitFromText(self, text: str):
        """
        Creates a new Unit object based on its text representation.
        
        Args:
            text (str): text representation of the new unit.
        
        Returns:
            object: unit object created from text.
        """
        pass
    
    @abstractmethod
    def unitToText(self, unit) -> str:
        """
        Returns the text representation of a unit object.
        
        Args:
            unit (object): unit object.
        
        Returns:
            str: text representation of the unit.
        """
        pass
    
    @abstractmethod
    def getQuantityValueNumeric(self, quantity):
        """Returns the numeric part of a quantity object.
        
        Args:
            quantity (object): quantity object.
        
        Returns:
            object: numeric representation of the quantity.
        """
        pass

    @abstractmethod
    def getQuantityValueStr(self, quantity) -> str:
        """Returns the numeric part of a quantity object converted to string.
        
        Args:
            quantity (object): quantity object.
        
        Returns:
            str: numeric part of the quantity converted to string.
        """
        pass
    
    @abstractmethod
    def getQuantityUnitStr(self, quantity) -> str:
        """
        Returns the text representation of the unit part of a
        quantity object.
        
        Args:
            quantity (object): quantity object.
        
        Returns:
            str: text representation of unit part.
        """
        pass
    
    @abstractmethod
    def getQuantityUnit(self, quantity):
        """
        Returns the unit object of a quantity.
        
        Args:
            quantity (object): quantity object.
        
        Returns:
            object: unit object of the quantity.
        """
        pass
    
    @abstractmethod
    def quantityTextRepr(self, quantity, unit_separator: str, normalize: bool, formatter) -> str:
        """
        Returns the text representation of a quantity object.
        
        Args:
            quantity (object): quantity object.
            unit_separator (str): separator between numeric and unit parts.
            normalize (bool): whether to normalize the numeric part.
            formatter (function): a function to be applied on the quantity that returns the text.
        
        Returns:
            str: text representation of the quantity.
        """
        pass

    @abstractmethod
    def isQuantityCompatibleWithUnit(self, quantity, unit) -> bool:
        """Verifies if a quantity is compatible with an unit.
        
        Args:
            quantity (object): quantity object.
            unit (Union[object, str]): unit object or its text representation.

        Returns:
            bool: True if it is compatible, False otherwise.
        """
        pass
    
    @abstractmethod
    def isUnitsCompatible(self, unit1, unit2) -> bool:
        """Verifies if two units are compatible.

        Args:
            unit1 (Union[object, str]): unit object 1 or its text representation.
            unit2 (Union[object, str]): unit object 2 or its text representation.

        Returns:
            bool: True if they are compatible, False otherwise.
        """
        pass

    @abstractmethod
    def isQuantitiesCompatible(self, q1, q2) -> bool:
        """
        Verifies if two quantities are compatible.

        Args:
            q1 (object): quantity object 1.
            q2 (object): quantity object 2.

        Returns:
            bool: True if the quantities are compatible, False otherwise.
        """
        pass
    
    @abstractmethod
    def isArrayOfSameDimension(self, array) -> bool:
        """
        Verifies if an array of units are all of the same dimension.

        Args:
            array (List[str]): array of unit strings

        Returns:
            bool: True if the array is of the same dimension, False otherwise.
        """
        pass
    
    @abstractmethod
    def isQuantitiesUnitsEqual(self, q1, q2) -> bool:
        """
        Verifies if the units of two quantities have the same text representations.

        Args:
            q1 (object): quantity object 1.
            q2 (object): quantity object 2.

        Returns:
            bool: True if the units have the same text representation, False otherwise.
        """
        pass
    
    def isUnitsEqual(self, u1, u2) -> bool:
        """
        Verifies if two units have equal text representations.

        Args:
            u1 (Union[object, str]): unit object 1 or its text representation.
            u2 (Union[object, str]): unit object 2 or its text representation.

        Returns:
            bool: True if their text representations are equal, False otherwise.
        """
        pass

    @abstractmethod
    def changeQuantityUnit(self, quantity, new_unit, formatter):
        """Returns a new quantity object with the new unit.
        
        Takes in a quantity object with numeric and unit parts,
        and returns a new quantity object with the numeric and
        unit parts converted to the new unit.

        Args:
            quantity (object): quantity object.
            new_unit (Union[object, str]): unit object or its text representation.
            formatter (function): a function to be applied on the quantity instead of the default conversion.
        
        Returns:
            object: new quantity object with the new unit and numeric parts.
        """
        pass

    @abstractmethod
    def __repr__(self) -> str:
        """Returns a string representation of the interface."""
        pass

class PintInterface(BackendInterface):
    """
    Interface between pint and ScientificSpinbox

    Implements required methods for creating quantities,
    verifying units and converting units.
    """

    def __init__(self, unit_system: str = 'SI', precision=30):
        """
        Args:
            unit_system (str): unit system to use on the pint.UnitRegistry.
                Defaults to 'SI'.
            precision (int): precision to use for Decimal. Defaults to 30.
        """
        self._unitSystem = unit_system
        self._unitRegistry = pint.UnitRegistry(
            system=self._unitSystem,
            case_sensitive=True,
            autoconvert_offset_to_baseunit=False,
        )
        self._unitRegistry.formatter.default_format = '~'
        self._precision = precision

        # Sets Decimal context
        self._decimalsContext = Context(
            prec=self._precision,       # maximum precision
            capitals=1,                 # prints exponential in uppercase
        )
        setcontext(self._decimalsContext)

    QuantityType = pint.Quantity
    ValueType = Decimal

    @property
    def decimalsContext(self):
        """
        Context to be used in Decimal library.
        """
        return self._decimalsContext

    @property
    def precision(self):
        """
        int: precision of the numeric representation.
        """
        return self._precision

    @property
    def unitSystem(self):
        """
        str: the unit system being used with pint.
        """
        return self._unitSystem

    @property
    def unitRegistry(self):
        """
        pint.UnitRegistry: the unit registry being used.
        """
        return self._unitRegistry

    def quantityFromText(self, text: str, unit: str):
        """
        Creates a new quantity object based on the text representation.

        Args:
            text (str): text representation of the new quantity.

        Returns:
            pint.Quantity: quantity object created from text.
        """
        return self.unitRegistry.Quantity(Decimal(text), unit)
    
    def quantityFromDecimal(self, value: Decimal, unit: str):
        """
        Creates a new quantity object based on the numeric representation.

        Args:
            value (Decimal): numeric representation of the new quantity.

        Returns:
            pint.Quantity: quantity object created from numeric representation.
        """
        return self.unitRegistry.Quantity(value, unit)

    def isUnitRegistered(self, unit: str) -> bool:
        """Verifies if a unit is registered in the Unit Registry.
        
        Args:
            unit (str): unit representation in text.

        Returns:
            bool: True if it is registered, False otherwise.
        """
        try:
            _ = self.unitRegistry.Unit(unit)
            return True
        except Exception:
            return False
        
    def unitFromText(self, text: str):
        """
        Creates a new Unit object based on its text representation.
        
        Args:
            text (str): text representation of the new unit.
            
        Raises:
            InvalidUnitError: if the unit is not registered in the interface's UnitRegistry.
        
        Returns:
            pint.Unit: unit object created from text.
        """
        if not self.isUnitRegistered(text):
            raise InvalidUnitError(text)
        return self.unitRegistry.Unit(text)
        
    def unitToText(self, unit: pint.Unit) -> str:
        """
        Returns the text representation of a unit object.
        
        Args:
            unit (pint.Unit): unit object.
        
        Returns:
            str: text representation of the unit.
        """
        return f"{unit:~}"
        
    def getQuantityValueNumeric(self, quantity: pint.Quantity) -> Decimal:
        """Returns the numeric part of a quantity object as Decimal.
        
        Args:
            quantity (pint.Quantity): quantity object.
        
        Returns:
            Decimal: numeric representation of the quantity.
        """
        return quantity.m

    def getQuantityValueStr(self, quantity: pint.Quantity) -> str:
        """Returns the numeric part of a quantity object converted to string.
        
        Args:
            quantity (pint.Quantity): quantity object.
        
        Returns:
            str: numeric part of the quantity converted to string.
        """
        return f'{quantity.m}'
    
    def getQuantityUnitStr(self, quantity: pint.Quantity) -> str:
        """
        Returns the text representation of the unit part of a
        quantity object.
        
        Args:
            quantity (pint.Quantity): quantity object.
        
        Returns:
            str: text representation of unit part.
        """
        unit = f'{quantity.u:~}'
        unit_splitted = list(unit)

        # Algorithm to replace possible multiplier symbols e.g. greek characters
        # with preferred characters.
        for symbol in multiplier_symbols.keys():
            item = multiplier_symbols[symbol]
            preferred = item["preferred"]
            for possible in item["possibles"]:
                if possible in unit_splitted:
                    unit = unit.replace(possible, preferred)
        return unit
    
    def getQuantityUnit(self, quantity: pint.Quantity) -> pint.Unit:
        """
        Returns the unit object of a quantity.
        
        Args:
            quantity (pint.Quantity): quantity object.
        
        Returns:
            pint.Unit: unit object of the quantity.
        """
        return quantity.u
    
    def quantityTextRepr(self,
                         quantity: pint.Quantity, 
                         unit_separator: str, 
                         normalize: bool = False, 
                         formatter = lambda x: f"{x:f}") -> str:
        """
        Returns the text representation of a quantity object.
        
        Args:
            quantity (pint.Quantity): quantity object.
            unit_separator (str): separator between numeric and unit parts.
            normalize (bool): whether to normalize the numeric part.
                Defaults to False.
            formatter (function): a function to be applied on the quantity that returns the text.
                Defaults to ``lambda x: f"{x:f}"``.
        
        Returns:
            str: text representation of the quantity.
        """
        if normalize:
            if formatter:
                return f'{formatter(quantity.m.normalize())}{unit_separator}{self.getQuantityUnitStr(quantity)}'
            else:
                return f'{quantity.m.normalize()}{unit_separator}{self.getQuantityUnitStr(quantity)}'
        else:
            if formatter:
                return f'{formatter(quantity.m)}{unit_separator}{self.getQuantityUnitStr(quantity)}'
            else:
                return f'{quantity.m}{unit_separator}{self.getQuantityUnitStr(quantity)}'

    def isQuantityCompatibleWithUnit(self, quantity: pint.Quantity, unit: Union[pint.Unit, str]) -> bool:
        """Verifies if a quantity is compatible with an unit.
        
        Args:
            quantity (pint.Quantity): quantity object.
            unit (Union[pint.Unit, str]): unit object or its text representation.

        Returns:
            bool: True if it is compatible, False otherwise.
        """
        if isinstance(unit, str) and not self.isUnitRegistered(unit):
            return False
        try:
            is_compatible = quantity.is_compatible_with(unit)
            return is_compatible
        except Exception:
            return False
            
    def isUnitsCompatible(self, unit1: str, unit2: str) -> bool:
        """Verifies if two units are compatible.

        Args:
            unit1 (Union[pint.Unit, str]): unit object 1 or its text representation.
            unit2 (Union[pint.Unit, str]): unit object 2 or its text representation.

        Returns:
            bool: True if they are compatible, False otherwise.
        """
        try:
            u1 = self.unitRegistry.Unit(unit1)
            u2 = self.unitRegistry.Unit(unit2)
            return u1.is_compatible_with(u2)
        except Exception:
            return False

    def isQuantitiesCompatible(self, q1: pint.Quantity, q2: pint.Quantity) -> bool:
        """
        Verifies if two quantities are compatible.

        Args:
            q1 (pint.Quantity): quantity object 1.
            q2 (pint.Quantity): quantity object 2.

        Returns:
            bool: True if the quantities are compatible, False otherwise.
        """
        try:
            is_compatible = q1.is_compatible_with(q2)
            return is_compatible
        except Exception:
            return False
        
    def isArrayOfSameDimension(self, array: List[str]) -> bool:
        """
        Verifies if an array of units are all of the same dimension.

        Args:
            array (List[str]): array of unit strings

        Returns:
            bool: True if the array is of the same dimension, False otherwise.
        """
        if len(array) < 1:
            raise EmptyArrayError()
        
        for item in array:
            if not self.isUnitRegistered(item):
                return False
            if not self.isUnitsCompatible(array[0], item):
                return False
        return True
    
    def isQuantitiesUnitsEqual(self, q1, q2) -> bool:
        """
        Verifies if the units of two quantities have the same text representations.

        Args:
            q1 (pint.Quantity): quantity object 1.
            q2 (pint.Quantity): quantity object 2.

        Returns:
            bool: True if the units have the same text representation, False otherwise.
        """
        if f"{self.getQuantityUnitStr(q1)}" == f"{self.getQuantityUnitStr(q2)}":
            return True
        else:
            return False
        
    def isUnitsEqual(self, u1, u2) -> bool:
        """
        Verifies if two units have equal text representations.

        Args:
            u1 (pint.Unit): unit object 1.
            u2 (pint.Unit): unit object 2.

        Returns:
            bool: True if their text representations are equal, False otherwise.
        """
        if f"{u1:~}" == f"{u2:~}":
            return True
        else:
            return False

    def changeQuantityUnit(self, quantity: pint.Quantity, new_unit: Union[str, pint.Unit], formatter=None):
        """Returns a new quantity object with the new unit.
        
        Takes in a quantity object with numeric and unit parts,
        and returns a new quantity object with the numeric and
        unit parts converted to the new unit.

        Args:
            quantity (pint.Quantity): quantity object.
            new_unit (Union[pint.Unit, str]): unit object or its text representation.
            formatter (function): a function to be applied on the quantity instead of the default conversion.
                If it is None, doesn't apply. Defaults to None.
        
        Returns:
            pint.Quantity: new quantity object with the new unit and numeric parts.
        """
        try:
            new_quantity = quantity.to(new_unit)
            if formatter:
                return formatter(quantity, new_quantity)
            else:
                return new_quantity
        except pint.DimensionalityError:
            raise IncompatibleConversionUnitError()

    def __repr__(self):
        """Returns a string representation of the PintInterface."""
        return f"PintInterface(unit_system=`{self.unitSystem}`, precision={self.precision})"


import sys
from decimal import Decimal

import pytest
from qtpy import QtCore, QtWidgets

from scientific_spinbox.widget import ScientificSpinBox
from scientific_spinbox.validators import ScientificInputValidator
from scientific_spinbox.defaults import set_default_interface
from scientific_spinbox.backend.interfaces import PintInterface

@pytest.fixture(scope='function', autouse=False)
def validator(qtbot):
    dummy_widget = QtWidgets.QLineEdit()
    qtbot.addWidget(dummy_widget)
    validator = ScientificInputValidator(parent=dummy_widget)
    dummy_widget.setValidator(validator)

    yield dummy_widget.validator()
    validator.deleteLater()

@pytest.fixture(scope='package', autouse=False)
def backend():
    interface = PintInterface('SI')
    set_default_interface(interface)
    return interface

@pytest.fixture(scope='function', autouse=False)
def gui_utils(qtbot):
    class GUIUtils:
        def __init__(self):
            pass

        def eraseAndEnterText(self, spinbox, text):
            linedit = spinbox.lineEdit()
            linedit.clear()
            qtbot.keyClicks(linedit, text)
            qtbot.keyClick(linedit, QtCore.Qt.Key_Enter)
    
    obj = GUIUtils()
    yield obj

@pytest.fixture(scope='function', autouse=False)
def widget_default(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()
    
@pytest.fixture(scope='function', autouse=False)
def widget_dimensionless_wo_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        thousand_separators=False,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def widget_dimensionless_with_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        thousand_separators=True,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def widget_with_baseunit_wo_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        base_unit='m',
        thousand_separators=False,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def widget_with_baseunit_with_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        base_unit='m',
        thousand_separators=True,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def widget_with_displayunit_with_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        base_unit="m",
        display_unit='m',
        thousand_separators=True,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def widget_with_displayunit_wo_thousand_sep(qtbot, backend):
    spinbox = ScientificSpinBox(
        backend_interface = backend,
        base_unit="m",
        display_unit='m',
        thousand_separators=False,
    )
    qtbot.addWidget(spinbox)
    yield spinbox
    spinbox.deleteLater()

@pytest.fixture(scope='function', autouse=False)
def all_widgets(qtbot,
                widget_dimensionless_wo_thousand_sep,
                widget_dimensionless_with_thousand_sep,
                widget_with_baseunit_wo_thousand_sep,
                widget_with_baseunit_with_thousand_sep,
                widget_with_displayunit_with_thousand_sep,
                widget_with_displayunit_wo_thousand_sep
    ):
    
    obj = {
        'dimensionless_wo_thousand_sep': widget_dimensionless_wo_thousand_sep,
        'dimensionless_with_thousand_sep': widget_dimensionless_with_thousand_sep,
        'with_baseunit_wo_thousand_sep': widget_with_baseunit_wo_thousand_sep,
        'with_baseunit_with_thousand_sep': widget_with_baseunit_with_thousand_sep,
        'with_displayunit_with_thousand_sep': widget_with_displayunit_with_thousand_sep,
        'with_displayunit_wo_thousand_sep': widget_with_displayunit_wo_thousand_sep
    }

    for key in obj.keys():
        qtbot.addWidget(obj[key])
    
    yield obj

    for key in obj.keys():
        obj[key].deleteLater()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'dlg_testspinbox_shortexample_gui.ui'
#
# Created by: qtpy UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from qtpy import QtCore, QtGui, QtWidgets  # noqa: E402, F401


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(400, 255)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.groupBox = QtWidgets.QGroupBox(Dialog)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.valueLabel = QtWidgets.QLabel(self.groupBox)
        self.valueLabel.setObjectName("valueLabel")
        self.verticalLayout.addWidget(self.valueLabel)
        self.unitLabel = QtWidgets.QLabel(self.groupBox)
        self.unitLabel.setObjectName("unitLabel")
        self.verticalLayout.addWidget(self.unitLabel)
        self.baseValueLabel = QtWidgets.QLabel(self.groupBox)
        self.baseValueLabel.setObjectName("baseValueLabel")
        self.verticalLayout.addWidget(self.baseValueLabel)
        self.baseUnitLabel = QtWidgets.QLabel(self.groupBox)
        self.baseUnitLabel.setObjectName("baseUnitLabel")
        self.verticalLayout.addWidget(self.baseUnitLabel)
        self.gridLayout.addWidget(self.groupBox, 4, 0, 1, 1)
        self.label_8 = QtWidgets.QLabel(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_8.sizePolicy().hasHeightForWidth())
        self.label_8.setSizePolicy(sizePolicy)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 0, 0, 1, 1)
        self.testSpinBox1 = QtWidgets.QWidget(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.testSpinBox1.sizePolicy().hasHeightForWidth())
        self.testSpinBox1.setSizePolicy(sizePolicy)
        self.testSpinBox1.setMinimumSize(QtCore.QSize(150, 30))
        self.testSpinBox1.setMaximumSize(QtCore.QSize(250, 30))
        self.testSpinBox1.setObjectName("testSpinBox1")
        self.gridLayout.addWidget(self.testSpinBox1, 2, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 50, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.gridLayout.addItem(spacerItem, 3, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 5, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.gridLayout.addItem(spacerItem1, 5, 0, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Test ScientificSpinBox"))
        self.groupBox.setTitle(_translate("Dialog", "Properties of interest"))
        self.valueLabel.setText(_translate("Dialog", "value:"))
        self.unitLabel.setText(_translate("Dialog", "unit:"))
        self.baseValueLabel.setText(_translate("Dialog", "baseValue:"))
        self.baseUnitLabel.setText(_translate("Dialog", "baseUnit:"))
        self.label_8.setText(_translate("Dialog", "ScientificSpinBox widget (controls time)"))



import pytest

from scientific_spinbox.backend.interfaces import PintInterface
from scientific_spinbox.defaults import set_default_interface, get_default_interface


def test_get_default_interface(backend):
    dummy_interface = PintInterface('SI')

    assert id(get_default_interface()) == id(backend)
    assert id(get_default_interface()) != id(dummy_interface)

def test_set_default_interface(backend):
    new_interface = PintInterface('SI')
    old_interface = get_default_interface()

    set_default_interface(new_interface)

    assert id(get_default_interface()) == id(new_interface)
    assert id(get_default_interface()) != id(old_interface)


# ScientificSpinBox errors
class InvalidBackendError(Exception):
    """Exception raised when backend interface is invalid."""
    def __init__(self):
        self._message = "ScientificSpinBox requires a valid backend interface object. Please pass it as a paremeter or set a valid default interface."
        super().__init__(self._message)

class NullBaseUnitError(Exception):
    """Exception raised when base unit is null."""
    def __init__(self):
        self._message = "`base_unit` must be provided."
        super().__init__(self._message)

class IncompatibleDisplayAndBaseUnitsError(Exception):
    """Exception raised when display and base units are incompatible."""
    def __init__(self):
        self._message = "`display_unit` and `base_unit` must have the same dimension."
        super().__init__(self._message)

class ArrayWithIncompatibleUnitsError(Exception):
    """Exception raised when an array with incompatible units is provided."""
    def __init__(self):
        self._message = "all units in the array must have the same dimension."
        super().__init__(self._message)

class ArrayIncompatibleWithBaseUnitError(Exception):
    """Exception raised when an array with units incompatible with base is provided."""
    def __init__(self):
        self._message = "`allowed_units` and `base_unit` must have the same dimension."
        super().__init__(self._message)

class DisplayUnitNotInAllowedUnitsError(Exception):
    """Exception raised when display unit is not in allowed units."""
    def __init__(self):
        self._message = "`display_unit` must be in `allowed_units`."
        super().__init__(self._message)

class BaseUnitNotInAllowedUnitsError(Exception):
    """Exception raised when base unit is not in allowed units."""

    def __init__(self):
        self._message = "`base_unit` must be in `allowed_units`."
        super().__init__(self._message)

class UnallowedFeatureError(Exception):
    """Exception raised when an unallowed feature is used."""
    def __init__(self, message: str):
        self._message = message
        super().__init__(self._message)


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'dlg_testspinbox_gui.ui'
#
# Created by: qtpy UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from qtpy import QtCore, QtGui, QtWidgets  # noqa: E402, F401


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(1037, 637)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox_2 = QtWidgets.QGroupBox(Dialog)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.testSpinBox1 = QtWidgets.QWidget(self.groupBox_2)
        self.testSpinBox1.setObjectName("testSpinBox1")
        self.verticalLayout_4.addWidget(self.testSpinBox1)
        self.reprPropertiesGroupBox = QtWidgets.QGroupBox(self.groupBox_2)
        self.reprPropertiesGroupBox.setObjectName("reprPropertiesGroupBox")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.reprPropertiesGroupBox)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.reprQuantityLabel = QtWidgets.QLabel(self.reprPropertiesGroupBox)
        self.reprQuantityLabel.setObjectName("reprQuantityLabel")
        self.gridLayout_5.addWidget(self.reprQuantityLabel, 1, 0, 1, 1)
        self.reprValueLabel = QtWidgets.QLabel(self.reprPropertiesGroupBox)
        self.reprValueLabel.setObjectName("reprValueLabel")
        self.gridLayout_5.addWidget(self.reprValueLabel, 0, 0, 1, 1)
        self.reprUnitLabel = QtWidgets.QLabel(self.reprPropertiesGroupBox)
        self.reprUnitLabel.setObjectName("reprUnitLabel")
        self.gridLayout_5.addWidget(self.reprUnitLabel, 2, 0, 1, 1)
        self.verticalLayout_4.addWidget(self.reprPropertiesGroupBox)
        self.groupBox_5 = QtWidgets.QGroupBox(self.groupBox_2)
        self.groupBox_5.setObjectName("groupBox_5")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_5)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.baseValueLabel = QtWidgets.QLabel(self.groupBox_5)
        self.baseValueLabel.setObjectName("baseValueLabel")
        self.verticalLayout_3.addWidget(self.baseValueLabel)
        self.baseQuantityLabel = QtWidgets.QLabel(self.groupBox_5)
        self.baseQuantityLabel.setObjectName("baseQuantityLabel")
        self.verticalLayout_3.addWidget(self.baseQuantityLabel)
        self.baseUnitLabel = QtWidgets.QLabel(self.groupBox_5)
        self.baseUnitLabel.setObjectName("baseUnitLabel")
        self.verticalLayout_3.addWidget(self.baseUnitLabel)
        self.verticalLayout_4.addWidget(self.groupBox_5)
        self.groupBox_6 = QtWidgets.QGroupBox(self.groupBox_2)
        self.groupBox_6.setObjectName("groupBox_6")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_6)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.internalTextLabel = QtWidgets.QLabel(self.groupBox_6)
        self.internalTextLabel.setObjectName("internalTextLabel")
        self.verticalLayout_2.addWidget(self.internalTextLabel)
        self.decimalsLabel = QtWidgets.QLabel(self.groupBox_6)
        self.decimalsLabel.setObjectName("decimalsLabel")
        self.verticalLayout_2.addWidget(self.decimalsLabel)
        self.lastTextLabel = QtWidgets.QLabel(self.groupBox_6)
        self.lastTextLabel.setObjectName("lastTextLabel")
        self.verticalLayout_2.addWidget(self.lastTextLabel)
        self.verticalLayout_4.addWidget(self.groupBox_6)
        self.verticalLayout.addWidget(self.groupBox_2)
        spacerItem = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)
        self.groupBox = QtWidgets.QGroupBox(Dialog)
        self.groupBox.setAutoFillBackground(False)
        self.groupBox.setFlat(False)
        self.groupBox.setCheckable(False)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.groupBox_3 = QtWidgets.QGroupBox(self.groupBox)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.tooltipCheckBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.tooltipCheckBox.setChecked(True)
        self.tooltipCheckBox.setObjectName("tooltipCheckBox")
        self.gridLayout_2.addWidget(self.tooltipCheckBox, 1, 0, 1, 1)
        self.thousandSepCheckBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.thousandSepCheckBox.setChecked(True)
        self.thousandSepCheckBox.setObjectName("thousandSepCheckBox")
        self.gridLayout_2.addWidget(self.thousandSepCheckBox, 0, 0, 1, 1)
        self.integerRoundCheckBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.integerRoundCheckBox.setEnabled(True)
        self.integerRoundCheckBox.setCheckable(True)
        self.integerRoundCheckBox.setObjectName("integerRoundCheckBox")
        self.gridLayout_2.addWidget(self.integerRoundCheckBox, 2, 0, 1, 1)
        self.gridLayout_3.addWidget(self.groupBox_3, 0, 0, 1, 1)
        self.applyButton = QtWidgets.QPushButton(self.groupBox)
        self.applyButton.setAutoDefault(False)
        self.applyButton.setObjectName("applyButton")
        self.gridLayout_3.addWidget(self.applyButton, 1, 0, 1, 2)
        self.groupBox_4 = QtWidgets.QGroupBox(self.groupBox)
        self.groupBox_4.setObjectName("groupBox_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_4)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label = QtWidgets.QLabel(self.groupBox_4)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.baseUnitTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.baseUnitTextEdit.sizePolicy().hasHeightForWidth())
        self.baseUnitTextEdit.setSizePolicy(sizePolicy)
        self.baseUnitTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.baseUnitTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.baseUnitTextEdit.setObjectName("baseUnitTextEdit")
        self.gridLayout.addWidget(self.baseUnitTextEdit, 0, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.groupBox_4)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.displayUnitTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.displayUnitTextEdit.sizePolicy().hasHeightForWidth())
        self.displayUnitTextEdit.setSizePolicy(sizePolicy)
        self.displayUnitTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.displayUnitTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.displayUnitTextEdit.setObjectName("displayUnitTextEdit")
        self.gridLayout.addWidget(self.displayUnitTextEdit, 1, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.groupBox_4)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)
        self.defaultValueTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.defaultValueTextEdit.sizePolicy().hasHeightForWidth())
        self.defaultValueTextEdit.setSizePolicy(sizePolicy)
        self.defaultValueTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.defaultValueTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.defaultValueTextEdit.setObjectName("defaultValueTextEdit")
        self.gridLayout.addWidget(self.defaultValueTextEdit, 2, 1, 1, 1)
        self.horizontalLayout.addLayout(self.gridLayout)
        spacerItem1 = QtWidgets.QSpacerItem(90, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_4 = QtWidgets.QLabel(self.groupBox_4)
        self.label_4.setObjectName("label_4")
        self.gridLayout_4.addWidget(self.label_4, 0, 0, 1, 1)
        self.defaultDecimalsTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.defaultDecimalsTextEdit.sizePolicy().hasHeightForWidth())
        self.defaultDecimalsTextEdit.setSizePolicy(sizePolicy)
        self.defaultDecimalsTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.defaultDecimalsTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.defaultDecimalsTextEdit.setObjectName("defaultDecimalsTextEdit")
        self.gridLayout_4.addWidget(self.defaultDecimalsTextEdit, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.groupBox_4)
        self.label_5.setObjectName("label_5")
        self.gridLayout_4.addWidget(self.label_5, 1, 0, 1, 1)
        self.allowedUnitsTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.allowedUnitsTextEdit.sizePolicy().hasHeightForWidth())
        self.allowedUnitsTextEdit.setSizePolicy(sizePolicy)
        self.allowedUnitsTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.allowedUnitsTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.allowedUnitsTextEdit.setObjectName("allowedUnitsTextEdit")
        self.gridLayout_4.addWidget(self.allowedUnitsTextEdit, 1, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.groupBox_4)
        self.label_6.setObjectName("label_6")
        self.gridLayout_4.addWidget(self.label_6, 2, 0, 1, 1)
        self.preferredUnitsTextEdit = QtWidgets.QTextEdit(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.preferredUnitsTextEdit.sizePolicy().hasHeightForWidth())
        self.preferredUnitsTextEdit.setSizePolicy(sizePolicy)
        self.preferredUnitsTextEdit.setMinimumSize(QtCore.QSize(0, 25))
        self.preferredUnitsTextEdit.setMaximumSize(QtCore.QSize(16777215, 25))
        self.preferredUnitsTextEdit.setObjectName("preferredUnitsTextEdit")
        self.gridLayout_4.addWidget(self.preferredUnitsTextEdit, 2, 1, 1, 1)
        self.horizontalLayout.addLayout(self.gridLayout_4)
        self.gridLayout_3.addWidget(self.groupBox_4, 0, 1, 1, 1)
        self.verticalLayout.addWidget(self.groupBox)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setEnabled(True)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.NoButton)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.groupBox_2.setTitle(_translate("Dialog", "Test ScientificSpinBox"))
        self.reprPropertiesGroupBox.setTitle(_translate("Dialog", "representation Properties"))
        self.reprQuantityLabel.setText(_translate("Dialog", "quantity: "))
        self.reprValueLabel.setText(_translate("Dialog", "value:"))
        self.reprUnitLabel.setText(_translate("Dialog", "unit: "))
        self.groupBox_5.setTitle(_translate("Dialog", "base Properties"))
        self.baseValueLabel.setText(_translate("Dialog", "baseValue:"))
        self.baseQuantityLabel.setText(_translate("Dialog", "baseQuantity:"))
        self.baseUnitLabel.setText(_translate("Dialog", "baseUnit:"))
        self.groupBox_6.setTitle(_translate("Dialog", "internal Properties"))
        self.internalTextLabel.setText(_translate("Dialog", "internalText:"))
        self.decimalsLabel.setText(_translate("Dialog", "decimals:"))
        self.lastTextLabel.setText(_translate("Dialog", "lastText:"))
        self.groupBox.setTitle(_translate("Dialog", "Controls"))
        self.groupBox_3.setTitle(_translate("Dialog", "Flags"))
        self.tooltipCheckBox.setText(_translate("Dialog", "Enable tooltip"))
        self.thousandSepCheckBox.setText(_translate("Dialog", "Enable thousand separators"))
        self.integerRoundCheckBox.setText(_translate("Dialog", "Enable rounding to next integer"))
        self.applyButton.setText(_translate("Dialog", "Apply changes"))
        self.groupBox_4.setTitle(_translate("Dialog", "Unit handling"))
        self.label.setText(_translate("Dialog", "Base unit:"))
        self.baseUnitTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.label_2.setText(_translate("Dialog", "Display unit:"))
        self.displayUnitTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.label_3.setText(_translate("Dialog", "Default value:"))
        self.defaultValueTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.label_4.setText(_translate("Dialog", "Default decimals:"))
        self.defaultDecimalsTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.label_5.setText(_translate("Dialog", "Allowed units:"))
        self.allowedUnitsTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))
        self.label_6.setText(_translate("Dialog", "Preferred units"))
        self.preferredUnitsTextEdit.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p></body></html>"))



# -*- coding: utf-8 -*-
"""
Widget module.

Provides the widget and implements all logic related to
the User Interface (UI), graphical elements, and related 
resources.

Since:
  2024/02/06

Authors:
  - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""

from decimal import Decimal, setcontext, ROUND_DOWN
from typing import List
import logging
import os

from qtpy import QtCore, QtWidgets, QtGui

from .backend.interfaces import PintInterface
from .backend.utils import convert_to_nearest_preferred_unit, quantityChangeUnitFormatter
from .validators import ScientificInputValidator
from .errors.widget import (
    InvalidBackendError,
    NullBaseUnitError,
    IncompatibleDisplayAndBaseUnitsError,
    ArrayWithIncompatibleUnitsError,
    ArrayIncompatibleWithBaseUnitError,
    BaseUnitNotInAllowedUnitsError,
    DisplayUnitNotInAllowedUnitsError,
    UnallowedFeatureError
)

from .defaults import (
    THOUSAND_SEPARATOR,
    UNIT_SEPARATOR,
    get_default_interface
)

_logger = logging.getLogger(__name__)

IS_DEBUG_MODE = bool(os.getenv("SCIENTIFIC_SPINBOX_DEBUG_MODE", False))
DEBUG_SHOW_STACK = bool(os.getenv("SCIENTIFIC_SPINBOX_DEBUG_SHOW_STACK", False))

if IS_DEBUG_MODE:
    # only for debugging
    import inspect 
    
class ScientificSpinBox(QtWidgets.QDoubleSpinBox):
    """A Qt5 Widget to manipulate physical quantities with ease.

    Note:
        - When working with units, an unit being '' means it's dimensionless.
        - The ``baseQuantityChanged`` and ``baseValueChanged`` signals can be
          used to track changes in the base unit. They are emitted after editing
          is finished.
        - The ``valueChanged`` signal is only emitted after editing is finished.
        - The ``textChanged`` signal is only emitted after editing is finished.

    Todo:
        - implement the use of self._internalDecimalSeparator = '.'
    """

    # Creates custom signals
    baseQuantityChanged = QtCore.Signal(PintInterface.QuantityType)
    """
    pint.Quantity: Signal emitted when the base quantity is changed. Returns the new base quantity.
    """
    
    baseValueChanged = QtCore.Signal(PintInterface.ValueType)
    """
    Decimal: Signal emitted when the base value is changed. Returns the new base value.
    """
    
    displayUnitChanged = QtCore.Signal(str)
    """
    str: Signal emitted when the display unit is changed. Returns the new display unit.
    """

    decimalsChanged = QtCore.Signal(int)
    """
    int: Signal emitted when the decimals is changed. Returns the decimals.
    """

    def __init__(
            self,
            *args,
            thousand_separators: bool = True,
            default_decimals: int = 2,
            default_value: PintInterface.ValueType = 0,
            allowed_units: List[str] | str = None,
            preferred_units: List[str] | str = None,
            display_unit: str = '',
            enable_tooltip: bool = True,
            base_unit: str = '',
            backend_interface : PintInterface = None,
            **kwargs,
        ):
        """
        Args:
            thousand_separators (bool): if True, uses thousand separators when
                showing the text. Defaults to False.
            default_decimals (int): number of decimals to show by default.
                Defaults to 2.
            default_value (Decimal): default value of the widget.
                Defaults to 0.
            allowed_units (List[str]): list of allowed units. If provided, enables
                the 'allowed units' feature. Can't be used with 'preferred units' feature.
                Defaults to None.
            preferred_units (List[str]): list of preferred units. If provided, enables
                the `preferred units` feature. Can't be used with `allowed units` feature.
                Defaults to None.
            display_unit (str): unit which will be used to display on the spinbox. If
                using the 'preferred units' or 'allowed units' features, the features
                units array must contain the display_unit.
                Defaults to ``''`` (dimensionless).
            enable_tooltip (bool): if False, disables showing tooltips when error occurs.
                Defaults to True.
            base_unit (str): base unit in which the value will be stored on baseQuantity property.
                If base_unit='', allows only dimensionless values.
                Defaults to ``''`` (dimensionless).
            backend_interface (PintInterface): the backend interface object to use. 
                Defaults to None.
        """
        
        super().__init__(*args, **kwargs)
        
        # Connect debug signals
        if IS_DEBUG_MODE:
            self.valueChanged.connect(self.debugValueChanged)
            self.lineEdit().textChanged.connect(self.debugLineEditTextChanged)
            self.textChanged.connect(self.debugScientificSpinboxTextChanged)
            self.editingFinished.connect(self.debugEditingFinished)
            self.displayUnitChanged.connect(self.debugDisplayUnitChanged)
            self.decimalsChanged.connect(self.debugDecimalsChanged)
            self.baseValueChanged.connect(self.debugBaseValueChanged)
            self.baseQuantityChanged.connect(self.debugBaseQuantityChanged)

        # Handle arguments
        self.handleArgs(
            thousand_separators,
            default_decimals,
            default_value,
            allowed_units,
            preferred_units,
            display_unit,
            enable_tooltip,
            base_unit,
            backend_interface
        )

        self.mockedQuantityChangeUnitFormatter = lambda old_quantity, new_quantity: quantityChangeUnitFormatter(
            self._backend,
            self._inputValidator,
            self._defaultDecimals,
            UNIT_SEPARATOR,
            old_quantity,
            new_quantity
        )

        # Sets up the defaults and styles
        self.setupWidget()

    def handleArgs(
            self,
            thousand_separators,
            default_decimals,
            default_value,
            allowed_units,
            preferred_units,
            display_unit,
            enable_tooltip,
            base_unit,
            backend_interface,
        ):
        """
        Processes the arguments passed to the constructor.
        
        Do the necessary sanity checks and raises exceptions if needed.
        
        Args:
            thousand_separators (bool): if True, uses thousand separators when
                showing the text.
            default_decimals (int): number of decimals to show by default.
            default_value (Decimal): default value of the widget.
            allowed_units (List[str]): list of allowed units. If provided, enables
                the 'allowed units' feature. Can't be used with 'preferred units' feature.
            preferred_units (List[str]): list of preferred units. If provided, enables
                the `preferred units` feature. Can't be used with `allowed units` feature.
            display_unit (str): unit which will be used to display on the spinbox. If
                using the 'preferred units' or 'allowed units' features, the features
                units array must contain the display_unit.
            enable_tooltip (bool): if False, disables showing tooltips when error occurs.
                Defaults to True.
            base_unit (str): base unit in which the value will be stored on baseQuantity property.
                If base_unit='', allows only dimensionless values.
            backend_interface (PintInterface): the backend interface object to use.
        """
        self._backend = backend_interface if backend_interface is not None else get_default_interface()
        self._precision = self._backend.precision

        if self._backend is None:
            raise InvalidBackendError()

        self._enableDimensionless = False
        self._enableDisplayUnit = False
        self._enableAllowedUnits = False
        self._enablePreferredUnits = False

        if base_unit is None:
            raise NullBaseUnitError()
        
        if base_unit == '':
            self._enableDimensionless = True

        if display_unit is not None and display_unit != '':
            self._enableDisplayUnit = True

        if self._enableDisplayUnit:
            if not self._backend.isUnitsCompatible(display_unit, base_unit):
                raise IncompatibleDisplayAndBaseUnitsError()

        if allowed_units and preferred_units:
            raise UnallowedFeatureError('`allowed_units` and `preferred_units` are mutually exclusive.')
        
        if preferred_units and self._enableDisplayUnit:
            raise UnallowedFeatureError('`preferred_units` and `display_unit` are mutually exclusive.')
            
        if allowed_units:
            if not self._backend.isArrayOfSameDimension(allowed_units):
                raise ArrayWithIncompatibleUnitsError()
            if not self._backend.isUnitsCompatible(base_unit, allowed_units[0]):
                raise ArrayIncompatibleWithBaseUnitError()
            if base_unit not in allowed_units:
                raise BaseUnitNotInAllowedUnitsError()
            if self._enableDisplayUnit and display_unit not in allowed_units:
                raise DisplayUnitNotInAllowedUnitsError()

            self._enableAllowedUnits = True
            self._allowedUnits = allowed_units
        
        self._preferredUnits = None
        if preferred_units:
            if not self._backend.isArrayOfSameDimension(preferred_units):
                raise ArrayWithIncompatibleUnitsError()
            if not self._backend.isUnitsCompatible(base_unit, preferred_units[0]):
                raise ArrayIncompatibleWithBaseUnitError()
            
            # We can't use displayUnit and preferredUnits at the same time.
            self._enableDisplayUnit = False

            self._enablePreferredUnits = True
            self._preferredUnits = preferred_units

        # Unit handling variables
        self._baseUnit = base_unit
        self._displayUnit = display_unit
        self._enableTooltip = enable_tooltip
        self._inputValidator = ScientificInputValidator(self)

        # Value handling variables
        self._enableThousandSeparators = thousand_separators
        self._defaultDecimals = default_decimals
        self._decimals = default_decimals
        self._defaultValue = default_value
        self._minRange, self._maxRange = -1e+50, +1e+50

    def setupWidget(self):
        """
        Sets up the widget defaults before running
        """
        # Internal defaults
        self._internalUnitSeparator = '#'
        self._internalThousandSeparator = '_'
        self._internalText = ''
        self._lastCursorPosition = 0

        # Sets context for Decimal
        setcontext(self._backend.decimalsContext)
        
        # Set up unit handling defaults
        if self._enableDisplayUnit:
            self._lastText = f"{self._defaultValue:.{self._defaultDecimals}f}" + self._displayUnit
            self._quantity = self._backend.quantityFromText(
                f"{Decimal(self._defaultValue):.{self._defaultDecimals}f}",
                self._displayUnit
            )
            self._baseQuantity = self._backend.changeQuantityUnit(self._quantity, self._baseUnit)
        else:
            self._lastText = f"{self._defaultValue:.{self._defaultDecimals}f}" + self._baseUnit
            self._quantity = self._backend.quantityFromText(
                f"{Decimal(self._defaultValue):.{self._defaultDecimals}f}",
                self.baseUnit
            )
            self._baseQuantity = self._quantity

        self._value = self._backend.getQuantityValueNumeric(self._quantity)
        self._lastBaseValue = self._backend.getQuantityValueNumeric(self._baseQuantity)
        self._lastDisplayUnit = self._displayUnit
        self._lastDisplayValue = self._value

        #_logger.debug(fAfter setting up internals: "{self._lastText=}, {self._value=}")

        # Set up DoubleSpinBox defaults
        self.blockSignals(True)
        self.lineEdit().blockSignals(True)
        self.lineEdit().setText(
            self.convertInternalToText(self._lastText)
        )
        self.setValue(Decimal(self._value))
        self.setRange(self._minRange, self._maxRange)
        self.lineEdit().setCursorPosition(self._lastCursorPosition)
        self.onFocus = False
        self.setKeyboardTracking(False)
        self.setLocale(QtCore.QLocale('English'))
    
        # Sets up the stylesheet to support lineEdit text color changing on input error
        self.lineEdit().setObjectName('ScientificSpinBox_LineEdit')
        default_stylesheet = \
        """
            QLineEdit#ScientificSpinBox_LineEdit[isInputError="true"] {
                color: red;
                padding-top: auto;
                padding-bottom: auto;
            }
            QLineEdit#ScientificSpinBox_LineEdit[isInputError="false"] {
                color: white;
                padding-top: auto;
                padding-bottom: auto;
            }
        """
        # Do not use the default stylesheet, instead let the developer handle it.
        #self.lineEdit().setStyleSheet(default_stylesheet)
        self.lineEdit().setProperty('isInputError', "false")
        self.clearErrorColor()
        self.blockSignals(False)
        self.lineEdit().blockSignals(False)

        # Connect signals
        self.editingFinished.connect(self.onEditingFinished)

    def debugEditingFinished(self) -> None:
        stack_names=[]
        if DEBUG_SHOW_STACK:
            stack = inspect.stack()
            stack_names = []
            for i in range(len(stack)):
                stack_names.append(stack[i].function)
            sig_block = self.signalsBlocked()
        _logger.debug(f"Intercepted event: emmited editingFinished")
        pass

    def debugValueChanged(self, value: str) -> None:
        stack_names=[]
        if DEBUG_SHOW_STACK:
            stack = inspect.stack()
            stack_names = []
            for i in range(len(stack)):
                stack_names.append(stack[i].function)
            sig_block = self.signalsBlocked()
        _logger.debug(f"Intercepted event: emmited valueChanged, {value=}")
        pass

    def debugScientificSpinboxTextChanged(self, text: str) -> None:
        stack_names=[]
        if DEBUG_SHOW_STACK:
            stack = inspect.stack()
            stack_names = []
            for i in range(len(stack)):
                stack_names.append(stack[i].function)
            sig_block = self.signalsBlocked()
        _logger.debug(f"Intercepted event: emmited scientificSpinbox.textChanged, {text=}")
        pass

    def debugDisplayUnitChanged(self, unit: str) -> None:
        stack_names=[]
        if DEBUG_SHOW_STACK:
            stack = inspect.stack()
            stack_names = []
            for i in range(len(stack)):
                stack_names.append(stack[i].function)
            sig_block = self.signalsBlocked()
        _logger.debug(f"Intercepted event: emmited displayUnitChanged, {unit=}")
        pass

    def debugBaseValueChanged(self, value: PintInterface.ValueType) -> None:
        stack_names=[]
        if DEBUG_SHOW_STACK:
            stack = inspect.stack()
            stack_names = []
            for i in range(len(stack)):
                stack_names.append(stack[i].function)
            sig_block = self.signalsBlocked()
        _logger.debug(f"Intercepted event: emmited baseValueChanged, {value=}")
        pass

    def debugBaseQuantityChanged(self, quantity: PintInterface.QuantityType) -> None:
        _logger.debug(f"Intercepted event: emmited baseQuantityChanged, {quantity=}")
        pass

    def debugLineEditTextChanged(self, text: str) -> None:
        _logger.debug(f"Intercepted event: emmited lineEdit.textChanged, {text=}")
        pass
    
    def debugDecimalsChanged(self, decimals: int) -> None:
        _logger.debug(f"Intercepted event: emmited decimalsChanged, {decimals=}")
        pass

    @property
    def quantity(self) -> PintInterface.QuantityType:
        """PintInterface.QuantityType: quantity being displayed."""
        return self._quantity
  
    @property
    def backend(self) -> PintInterface:
        """PintInterface: backend interface used by the widget."""
        return self._backend

    @property
    def value(self) -> PintInterface.ValueType:
        """PintInterface.ValueType: value of the quantity being displayed."""
        return self._value
    
    @property
    def unit(self) -> str:
        """str: unit of the quantity being displayed."""
        return self._backend.getQuantityUnitStr(self._quantity)
    
    @value.setter
    def value(self, value: PintInterface.ValueType) -> None:
        if value != self._value:
            self.setValue(value)

    @property
    def baseValue(self) -> PintInterface.ValueType:
        """PintInterface.ValueType: value of the quantity converted to the base unit."""
        self._baseQuantity = self._backend.changeQuantityUnit(self._quantity, self._baseUnit)
        return self._backend.getQuantityValueNumeric(self._baseQuantity)
    
    @property
    def baseUnit(self) -> str:
        """str: unit of the base quantity."""
        return self._baseUnit
    
    @property
    def baseQuantity(self) -> PintInterface.QuantityType:
        """PintInterface.QuantityType: base quantity with the base unit and value converted to base unit."""
        self._baseQuantity = self._backend.changeQuantityUnit(self._quantity, self._baseUnit)
        return self._baseQuantity
    
    @property
    def defaultDecimals(self) -> int:
        """int: default number of decimals"""
        return self._defaultDecimals
    
    @QtCore.Slot()
    def onEditingFinished(self) -> None:
        """Executes necessary actions after editing finished.
        
        Actions:
            - Corrects the cursor position.
            - Formats the text.
            - Updates the base quantity.
        """        
        self.blockSignals(True)
        self.correctDecimals()
        last_position = self._lastCursorPosition

        # Formats the text correctly
        captured = self._inputValidator.getCapturedGroups(self.removeSeparators(self._internalText))
        numeric = f"{Decimal(captured.numericRepresentation):.{self._decimals}f}"
        new_internal = self.addSeparators(numeric + captured.unitText)
        self._internalText = new_internal
        self._lastText = self.convertInternalToText(new_internal)
        self.lineEdit().setText(self._lastText)

        # Corrects the cursor position
        self.keepCursorPosition(last_position=last_position, correct=True, is_editing_finished=True)

        # Updates the base quantity
        base_quantity = self._backend.changeQuantityUnit(self._quantity, self._baseUnit)
        self._baseQuantity = base_quantity

        self.blockSignals(False)

        # Emit custom signals
        if self._lastBaseValue != self.baseValue:
            self._lastBaseValue = self.baseValue
            if not self.signalsBlocked():
                self.baseValueChanged.emit(self.baseValue)
                self.baseQuantityChanged.emit(base_quantity)

        if self._lastDisplayUnit != self.unit:
            self._lastDisplayUnit = self.unit
            if not self.signalsBlocked():
                self.displayUnitChanged.emit(self.unit)
        #_logger.debug(f"After editing finished: {self._internalText=}, {self.baseQuantity=}, {self.baseValue=}, {self.baseUnit=}")

    def setErrorColor(self) -> None:
        """Sets the lineEdit text color to red.
        
        Sets the isInputError property to true and
        updates the lineEdit style.
        """
        self.lineEdit().setProperty('isInputError', "true")
        self.lineEdit().style().unpolish(self.lineEdit())
        self.lineEdit().style().polish(self.lineEdit())
        self.lineEdit().update()
        
    def clearErrorColor(self) -> None:
        """Sets the lineEdit text color to black.
        
        Sets the isInputError property to false and
        updates the lineEdit style.
        """
        self.lineEdit().setProperty('isInputError', "false")
        self.lineEdit().style().unpolish(self.lineEdit())
        self.lineEdit().style().polish(self.lineEdit())
        self.lineEdit().update()

    def showTooltip(self, text: str) -> None:
        """Shows immediately a tooltip with given text.
        
        Args:
            text (str): text to show on the tooltip. Can be HTML.
        """
        self.setToolTipDuration(-1)
        QtWidgets.QToolTip.showText(self.mapToGlobal(self.rect().bottomLeft()), text)

    def emptyUnitEvent(self) -> None:
        """Triggers tooltip and error color when unit is empty.

        This function is called on validation.
        """
        self.setErrorColor()
        if self._enableTooltip: 
            self.showTooltip("Unit can't be empty")

    def invalidUnitEvent(self) -> None:
        """Triggers tooltip and error color when unit is invalid.

        This function is called on validation.
        """
        self.setErrorColor()
        if self._enableTooltip:
            self.showTooltip("Invalid unit")

    def incompatibleUnitEvent(self) -> None:
        """Triggers tooltip and error color when unit is incompatible.

        This function is called on validation.
        """
        self.setErrorColor()
        if self._enableTooltip:
            self.showTooltip("Unit with incompatible dimensionality")

    def unitNotEmptyEvent(self) -> None:
        """Triggers tooltip and error color when unit is not empty.

        This function is called on validation.
        """
        self.setErrorColor()
        if self._enableTooltip:
            self.showTooltip("Unit must be empty.")

    def unitNotAllowedEvent(self) -> None:
        """Triggers tooltip and error color when unit is not allowed.

        This function is called on validation.
        """
        self.setErrorColor()
        if self._enableTooltip:
            self.showTooltip("Allowed units are: " + ", ".join(self._allowedUnits) + '.')

    def clearToolTip(self) -> None:
        """Hides the tooltip."""
        QtWidgets.QToolTip.hideText()
        
    def signalsBlocked(self):
        """
        Verify if signals are blocked.
        
        Returns:
            bool: True if signals are blocked and False otherwise.
        """
        return super().signalsBlocked()
    
    def blockSignals(self, block: bool) -> None:
        """
        Blocks the signals.
        
        Args:
            block (bool): True to block the signals and False to unblock.
        """
        return super().blockSignals(block)

    def setDisplayUnit(self, new_unit: str):
        """Converts the display quantity to a new display unit and sets the value.
        
        Args:
            new_unit (str): The new display unit.
        """
        self._lastDisplayUnit = self._displayUnit
        
        if not self._enablePreferredUnits:
            if self._enableAllowedUnits and new_unit not in self._allowedUnits:
                return

            # Sets the new display unit
            self._displayUnit = new_unit
            
            # Sets the new value and corrects the decimals
            self._quantity = self._backend.changeQuantityUnit(self._quantity, new_unit)
            self._value = self._backend.getQuantityValueNumeric(self._quantity)

            # Corrects the decimals
            # (displayUnitConversor does the job of converting the baseValue decimals to displayValue decimals)
            self._internalText = self.displayUnitConversor(f"{self._value:f}", self._displayUnit)
            self._lastText = self.convertInternalToText(self._internalText)

            self.setValue(self.backend.getQuantityValueNumeric(self._quantity))
            if not self.signalsBlocked():
                if self._lastDisplayUnit != self._displayUnit:
                    self.displayUnitChanged.emit(self.unit)
            
        else:
            return

    def setBaseValue(self, value: PintInterface.ValueType | float, from_setvalue = False) -> None:
        """Sets the widget base value.

        If the value is a float, converts the passed value
        to backend's ValueType, using the current number of decimals.

        If signals are not blocked, emits baseValueChanged with the value
        converted to ValueType.

        Args:
            value (PintInterface.ValueType): the value to set.
            from_setvalue (bool): flag to know if the function was called by setValue to
                avoid infinite loops.
        """
        if isinstance(value, float):
            # Allows handling floats differently
            value = Decimal(value)
        
        old_basevalue = self.baseValue
        self._lastBaseValue = old_basevalue
        self._baseQuantity = self.backend.quantityFromDecimal(value, self.baseUnit)
        self._quantity = self.backend.changeQuantityUnit(self._baseQuantity, self.unit)
        display_value = self.backend.getQuantityValueNumeric(self._quantity)

        new_basevalue = self._backend.getQuantityValueNumeric(self._baseQuantity)
        if not self.signalsBlocked():
            if new_basevalue != old_basevalue:
                self.baseValueChanged.emit(new_basevalue)
                self.baseQuantityChanged.emit(self._baseQuantity)
                
        if not from_setvalue:
            self.setValue(Decimal(display_value))

    def setValue(self, value: PintInterface.ValueType | float) -> None:
        """Sets the widget value.

        If the value is a float, converts the passed value
        to backend's ValueType, using the current number of decimals.

        If signals are not blocked, emits valueChanged with the value
        converted to float.

        Args:
            value (PintInterface.ValueType): the value to set.
        """

        if isinstance(value, float):
            # Allows handling floats differently
            value = Decimal(f"{value:f}")

        #_logger.debug(f"quantity repr = {self.backend.quantityTextRepr(self._quantity, UNIT_SEPARATOR)}")
        self.blockSignals(True)
        # Truncates the value to the current number of decimals
        new_numeric_text = f"{value:.{self._decimals}f}"
        value = Decimal(new_numeric_text)
        old_value = self._value
        
        self._value = value
        self._quantity = self.backend.quantityFromDecimal(value, self.unit)
        self._baseQuantity = self.backend.changeQuantityUnit(self._quantity, self._baseUnit)
        
        # Sets the text
        if self.unit:
            new_numeric_text += self.unit

        self._internalText = self.addSeparators(new_numeric_text)
        self._lastText = self.convertInternalToText(self._internalText)
        self.lineEdit().blockSignals(True)
        self.lineEdit().setText(self._lastText)
        self.lineEdit().blockSignals(False)

        #_logger.debug(f"{self._internalText=}, {self._value=}")
        self.blockSignals(False)

        # Emit custom signals
        # We only verify after changing, to allow for setDecimals to change the value
        # e.g. 1.00 == 1.000 but we still want this change to happen if someone issues setDecimals(3)
        if value != old_value:
            if not self.signalsBlocked():
                self.valueChanged.emit(self._value)
        if self.lineEdit().text() != self._lastText:
            if not self.signalsBlocked():
                self.textChanged.emit(self._lastText)
        if self._lastBaseValue != self.baseValue:
            self._lastBaseValue = self.baseValue
            if not self.signalsBlocked():
                self.baseValueChanged.emit(self.baseValue)
                self.baseQuantityChanged.emit(self._baseQuantity)

    def text(self) -> str:
        """Returns the widget text."""
        return self.lineEdit().text()
    
    def cleanText(self) -> str:
        """Returns only the numeric part of the displayed text."""
        return self.getNumericText()
    
    def suffix(self) -> str:
        """Returns the suffix, that is, the unit text."""
        return self.unit

    def decimals(self) -> int:
        """Returns the current number of decimals.
        
        Returns:
            int: number of decimals.
        """
        return self._decimals
    
    def setDecimals(self, prec: int) -> None:
        """Sets the number of decimals.
        
        Args:
            prec (int): number of decimals.
        """
        if isinstance(prec, int):
            if prec != self._decimals:
                self._decimals = prec
                self.decimalsChanged.emit(prec)
                self.blockSignals(True)
                self.setValue(Decimal(self._value))
                self.blockSignals(False)
        else:
            raise TypeError('The `value` passed to setDecimals must be an integer.')

    def getNumericText(self) -> str:
        """Returns only the scientific representation of the widget's internal text."""
        text = self.removeSeparators(self._internalText)
        captured = self._inputValidator.getCapturedGroups(text)
        return captured.scientificRepresentation

    def correctDecimals(self, text=None):
        """
        Corrects the number of decimal places according to
        the text, adapting to the input's decimal places.

        Note:
            - If ``text`` is not provided, uses the current internal text
              of the widget.
            - If ``text`` is provided, its expected to only contain the
              scientific representation text, e.g. ``+1234.54544E-9``.

        Examples:
            1. text = 1.23E-5 => decimal_change = 5+2 = 7
            2. text = 25e+5 => decimal_change = self._decimals
            3. text = 25e-3 and default_decimals = 2 => decimal_change = self._decimals
            4. text = 3.251 and default_decimals = 2 => decimal_change = 3
            5. text = 3.1 and default_decimals = 2 => decimal_change = 2

        Args:
            text (str, optional): text to be processed. Defaults to None.
        """
        if text:
            value_str = text.lower()
        else:
            value_str = self.getNumericText().lower()

        decimal_change = self._decimals

        if 'e' in value_str:
            real, exp = value_str.split('e')
            if int(exp) < 0:
                # Counts how many zeros between decimal point and the exponent.
                decimal_change = abs(int(exp))

                if '.' in real:
                    # If it's a real number before exponent, like 1.23E-5, adds the length of the
                    # decimal part of the real number and the size of the exponent.
                    # e.g.: 1.23E-5 = 0.0000123 => decimal_change = 5 + 2 = 7
                    integer, decimal = real.split('.')
                    decimal_change += len(decimal)

        if '.' in value_str and 'e' not in value_str:
            # If it's a real number with integer and decimal part, without exponent.
            integer, decimal = value_str.split('.')
            if decimal:
                decimal_reversed = int(decimal[::-1])
                if decimal_reversed == 0 or decimal is None:
                    # If the decimal part can be represented as zero
                    # Here we choose to NOT truncate to self._defaultDecimals.
                    decimal_change = len(decimal)
                    if decimal_change < self._defaultDecimals:
                        decimal_change = self._defaultDecimals
                else:
                    # If the decimal part is non-zero
                    new_decimal = len(decimal)
                    if new_decimal > self._defaultDecimals:
                        decimal_change = new_decimal
                    else:
                        decimal_change = self._defaultDecimals
            else:
                # If number is like "0." or "12345678."
                decimal_change = self._defaultDecimals

        if '.' not in value_str and 'e' not in value_str:
            # If it's a pure integer without decimal or exponent.
            decimal_change = self._defaultDecimals

        if value_str == '0':
            # If it's purely zero
            decimal_change = self._defaultDecimals

        # Updates the decimals if there's a change
        if value_str and decimal_change:
            #_logger.debug(f"Setting decimals: {value_str=}, {decimal_change=}")
            self.setDecimals(decimal_change)

    def validate(self, text: str, position: int) -> ScientificInputValidator.State:
        """
        Validates the text input using two layers of validation.

        First, the text is validated by ScientificInputValidator to
        verify if the text is in the expected format for physical quantities.

        If the result of the first validator is Acceptable, does a second validation
        to verify if the unit provided is valid, compatible and not empty.

        Triggers the tooltip and error color hints.

        Args:
            text (str): text to validate.
            position (int): position of the cursor before validation.

        Returns:
            state (ScientificInputValidator.State): state of the validator.
            text (str): text after the validation.
            position (int): position of the cursor after validation.
        """
        #_logger.debug(f'Incoming text: {self.removeSeparators(text)}')
        text_wo_separators = self.removeSeparators(text)
        pre_state, pre_text, position = self._inputValidator.validate(
            text_wo_separators,
            position
        )
        
        #_logger.debug(f'After 1st validation: pre_state={self._inputValidator.stateToStr(pre_state)}, pre_text={pre_text}, position={position}')

        if pre_state == self._inputValidator.State.Intermediate or pre_state == self._inputValidator.State.Invalid:
            return pre_state, text, position
        
        captured = self._inputValidator.getCapturedGroups(text_wo_separators)
        int_num = captured.integerNumbers
        decimal_num = captured.decimalNumbers
        exp_num = captured.exponentialNumber
        unit = captured.unitText
        #_logger.debug(f"{unit=}")

        if not int_num and not decimal_num and not exp_num:
            pre_state = self._inputValidator.State.Intermediate
            return pre_state, text, position

        is_unit_empty = (not unit) and (self._baseUnit is not None)
        is_unit_invalid = not self.backend.isUnitRegistered(unit)
        is_unit_incompatible_with_base = not self._backend.isUnitsCompatible(unit, self._baseUnit)
        is_unit_not_allowed = (unit not in self._allowedUnits) if self._enableAllowedUnits else False

        #_logger.debug(f'Validation flags: {is_unit_empty=}, {is_unit_invalid=}, {is_unit_incompatible_with_base=}')

        if not self._enableDimensionless:
            if is_unit_empty:
                self.emptyUnitEvent()

            if is_unit_invalid:
                self.invalidUnitEvent()

            if is_unit_incompatible_with_base:
                self.incompatibleUnitEvent()

            if not is_unit_incompatible_with_base and is_unit_not_allowed:
                self.unitNotAllowedEvent()

            is_error = (
                is_unit_invalid or                  \
                is_unit_incompatible_with_base or   \
                is_unit_empty or                    \
                is_unit_not_allowed                 \
            )

            if is_error:
                pre_state = self._inputValidator.State.Intermediate
                #_logger.debug(f'Error on validation: unit invalid, empty or incompatible, {unit=}')
            else:
                self.clearToolTip()
                self.clearErrorColor()
        else:
            if not is_unit_empty:
                # Dimensionless is enabled, so it's expected that unit is empty or editing is not finished.
                pre_state = self._inputValidator.State.Intermediate
                self.unitNotEmptyEvent()
                #_logger.debug(f"Expected dimensionless, but unit's not empty: {unit=}, {is_unit_empty=}, state={self._inputValidator.stateToStr(pre_state)}, {pre_text=}")
            else:
                self.clearToolTip()
                self.clearErrorColor()

        #_logger.debug(f"After validation: (state, text, position) = ({self._inputValidator.stateToStr(pre_state)}, {text}, {position})")

        return pre_state, text, position
    
    def correctCursorPosition(self, pre_position, is_step_by: bool = False, is_editing_finished: bool = False):
        """Corrects the cursor position considering special cases.

        If the is_step_by flag is True, does the following:
            1. if it was at the right-side of an unit, decimal or thousand
                separator, moves the cursor to the left-side of separator.
            2. if the cursor was after the numeric part, moves it to the
                right-side of the last character of the numeric part.
            3. if the cursor is at the left-side of the numeric part, on
                the signal place, moves it to the right-side of the first
                character of the numeric part.

        If the is_step_by flag is False, does the following:
            1. if the cursor was at the right-side of the unit separator,
                moves it to the left-side of unit separator.
            2. if the cursor was at the right-side of the last character of
                the text, moves it to the right-side of the last character
                of the numeric part.
        
        Args:
            pre_position (int): position of the cursor before correction.
            is_step_by (bool, optional): if True, considers it was called from
                a stepBy and handles differently. Defaults to False.
            is_editing_finished (bool, optional): flag to know if the function was
                called inside onEditingFinished and handle it differently. 
                Defaults to False.

        Returns:
            int: new cursor position.
        """
        text_wo_thousand_sep = self.convertInternalToText(
            self.removeThousandSeparators(self._internalText)
        )

        if self._enableThousandSeparators:
            quant_thousand_sep = self._internalText.count(self._internalThousandSeparator)
        else:
            quant_thousand_sep = 0

        captured = self._inputValidator.getCapturedGroups(text_wo_thousand_sep)
        numerical_len_raw = len(captured.scientificRepresentation)
        integer_len_raw = len(captured.integerNumbers)
        signal = captured.numericSignal

        if not is_step_by:
            #_logger.debug(f"Before 'not step by' logic: {pre_position=}, last_index_of(internalText)={len(self._internalText)-1}")
            if pre_position > 0:
                if pre_position - 1 <= len(self._internalText)-1 and self._internalText[pre_position - 1] == self._internalUnitSeparator:
                    # Moves the cursor 1 unit to the left if the cursor was 1 unit after a unit separator
                    new_position = pre_position - 1
                    return new_position
                if pre_position > len(self._internalText) - 1:
                    # Moves cursor to the end of the numeric part if the cursor was after the text
                    new_position = quant_thousand_sep + numerical_len_raw
                    return new_position

            if pre_position == 0 and is_editing_finished:
                # Moves cursor to the end of the numeric part if the cursor was before the text
                new_position = quant_thousand_sep + numerical_len_raw
                return new_position
            
            return pre_position
        
        # All conditions below must only be applied if is_step_by = true.
        new_position = pre_position
        
        if pre_position == 0 and signal != '-':
            new_position = pre_position + 1
            return new_position
        
        if pre_position <= 2 and signal == '-':
            new_position += (2 - pre_position)
            return new_position

        if pre_position > numerical_len_raw + quant_thousand_sep:
            # Moves cursor to the end of the numeric part if the cursor was after the numeric part
            # Only runs if its called by a stepby.
            new_position = quant_thousand_sep + numerical_len_raw
            return new_position

        if pre_position > 0 and self._internalText[pre_position - 1] == self._internalThousandSeparator:
            # Moves the cursor 1 unit to the left if the cursor was 1 unit after a thousand separator
            # Only runs if its called by a stepby
            new_position = pre_position - 1
            return new_position
        
        if pre_position > 0 and self._internalText[pre_position - 1] == '.':
            new_position = pre_position - 1
            return new_position

        #_logger.debug(f"After all logic: {is_step_by=}, {pre_position=}, {new_position=}")
        return new_position
    
    def getNumberToChange(self, cursor_pos: int, text: str, step: float):
        """
        Gets the number to be added/subtracted to the value based on
        cursor position.

        Attention:
            - Only tested to work with abs(step) = 1 currently.
            - Can't handle exponentials.

        Args:
            cursor_pos (int): current cursor position.
            text (str): current text.
            step (float): step to take.
        """        
        # Correct cursor position accounting for thousand separators
        thousand_sep_before_cursor = 0
        text_before_cursor = text[:cursor_pos]
        thousand_sep_before_cursor = text_before_cursor.count(self._internalThousandSeparator)
    
        cursor_pos_wo_sep = cursor_pos - thousand_sep_before_cursor
        text_wo_sep = self.removeSeparators(text)
        
        text_wo_units = self._inputValidator.getCapturedGroups(text_wo_sep).scientificRepresentation

        point_pos = text_wo_sep.find('.')
        if point_pos != -1:
            if cursor_pos_wo_sep <= point_pos:
                return Decimal('1' + '0'*(len(text_wo_sep[:point_pos]) - cursor_pos_wo_sep - 1)) * step

            if cursor_pos_wo_sep > point_pos:
                return Decimal('0.' + '0'*(cursor_pos_wo_sep - len(text_wo_sep[:point_pos]) - 1) + '1') * step
        else:
            # In case there's no decimal point
            return Decimal('1' + '0'*(len(text_wo_units) - cursor_pos_wo_sep - 1)) * step

        return Decimal('0.0')

    def numericSignalChange(self, text_before: str, text_after: str):
        """
        Detects if the number changed signal.

        Args:
            text_before (str): text before the change.
            text_after (str): text after the change.

        Returns:
            float: None if can't determine the signal, 1 if changed from
                   negative to positive, -1 if changed from positive to negative.
        """
        text_before = self.removeSeparators(text_before)
        text_after = self.removeSeparators(text_after)

        if len(text_before) < 1 or len(text_after) < 1:
            return None
        
        if text_before[0] == '-' and text_after[0] != '-':
            return 1
        
        if text_before[0] != '-' and text_after[0] == '-':
            return -1
        
        return None

    def stepBy(self, step: float):
        """Handles the step by functionality.
        
        Changes cursor position using correctCursorPosition to avoid separators
        and add/subtract the number using getNumberToChange.

        Also corrects the cursor position if the number signal is changed.

        Side effects:
            - sets the text and value, propagating the step.

        Args:
            step (float): step to take.
        """
        curr_pos = self.lineEdit().cursorPosition()
        last_internal_text = self._internalText

        # Corrects cursor position before calculating step
        new_position = self.correctCursorPosition(curr_pos, is_step_by=True)
        self.lineEdit().setCursorPosition(new_position)

        number_to_change = self.getNumberToChange(new_position - 1 if new_position > 0 else 0, last_internal_text, step)
        new_value = self.value + number_to_change

        #_logger.debug(f'{last_internal_text=}, {number_to_change=}, {new_value=}, {curr_pos=}, {new_position=}, {self._decimals=}')

        # Handles the stepby manually (alter text and value) instead of calling super().stepBy(step)
        new_numeric_text = f"{new_value:.{self._decimals}f}"

        # Block internal signal. The value will only be set after editing finished.
        self.blockSignals(True)
        self.setValue(Decimal(new_numeric_text))
        self.blockSignals(False)

        self._internalText = self.addSeparators(new_numeric_text)
        if self.unit:
            self._internalText += self._internalUnitSeparator + self.unit
        self._lastText = self.convertInternalToText(self._internalText)
        self.lineEdit().setText(self._lastText)
        new_internal_text = self._internalText

        # Handles cursor position when numeric signal changes
        signal_change = self.numericSignalChange(last_internal_text, new_internal_text)
        if signal_change:
            if signal_change < 0:
                # Changed from + to -
                new_position += 1

            if signal_change > 0:
                # Changed from - to +
                new_position -= 1

        # Maintains cursor position after step 
        if new_position == 0:
            new_position += 1

        # Move the cursor if the size of the text changes
        size_change = len(new_internal_text) - len(last_internal_text)
        if size_change and signal_change is None:
            new_position += size_change
        
        self.lineEdit().setCursorPosition(new_position)

    def addSeparators(self, text):
        """Formats text as internal text with internal separators.

        Uses ScientificInputValidator to get named captured groups from
        the provided text.

        Args:
            text (str): current text.

        Returns:
            str: text with internal separators.
        """        
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(text)
        )

        signal = captured.numericSignal
        integers = captured.integerNumbers
        decimals = captured.decimalNumbers
        exponential = captured.exponentialRepresentation
        unit = captured.unitText

        # Uses Python's native thousand separator formatting
        if self._enableThousandSeparators:
            integers = f"{int(integers):_}"
        else:
            integers = f"{int(integers)}"
        self._quant_thousand_sep = integers.count('_')
        integers = integers.replace('_', self._internalThousandSeparator)
        unit_text = self._internalUnitSeparator + unit if unit else ''
        decimals_text = '.' + decimals if decimals else ''

        if self._enableDimensionless:
            unit_text = ''
        
        new_text = f"{signal}{integers}{decimals_text}{exponential}{unit_text}"
        return new_text
    
    def removeSeparators(self, text):
        """
        Removes all internal and representation separators from text.

        Args:
            text (str): text to remove separators.
        
        Returns:
            str: text without separators.
        """
        text = text.replace(self._internalUnitSeparator, '')
        text = text.replace(self._internalThousandSeparator, '')
        text = text.replace(THOUSAND_SEPARATOR, '')
        text = text.replace(UNIT_SEPARATOR, '')
        return text
    
    def removeThousandSeparators(self, text: str):
        """Remove thousand separators from text.
        
        Args:
            text (str): text to remove thousand separators.
        
        Returns:
            str: text without thousand separators
        """
        if self._enableThousandSeparators:
            return text.replace(self._internalThousandSeparator, '')
        return text
    
    def convertInternalToText(self, text: str) -> str:
        """
        Converts the internal representation to the user's representation.

        Args:
            text (str): text formatted in internal representation.

        Returns:
            str: text formatted in user representation.
        """
        text = text.replace(self._internalThousandSeparator, THOUSAND_SEPARATOR)
        text = text.replace(self._internalUnitSeparator, UNIT_SEPARATOR)
        return text
    
    def valueFromText(self, text: str) -> float:
        """
        Retrieves the value from text and sets the current value.

        This method uses the ScientificInputValidator's regex parser
        to get the named capture groups of the current text and parse
        them into a value.

        Side effects:
            - sets the representation quantity.

        Args:
            text (str): the current widget's text.
        
        Returns:
            float: the value retrieved from the text.
        """
        self._lastText = text
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(self.lineEdit().text())
        )

        numeric = captured.scientificRepresentation
        unit = captured.unitText

        if self._enableDimensionless:
            unit = ''

        self.correctDecimals(text=numeric)

        new_quantity = self._backend.quantityFromText(numeric, unit)
        self._quantity = new_quantity

        #_logger.debug(f"After processing: {numeric=}, {unit=}, {self._decimals=}")

        self.blockSignals(True)
        self.setValue(Decimal(numeric))
        self.blockSignals(False)
        
        return self._value
    
    def scientificToDecimal(self, scientific_text):
        """
        Converts text with scientific notation to a Decimal number and
        returns its text representation.

        Args:
            scientific_text (str): the input text.

        Returns:
            str: the text representation of the Decimal number.
        """
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(scientific_text)
        )
        numeric_repr = captured.numericRepresentation
        exp_num = captured.exponentialNumber
        if not exp_num:
            return f"{Decimal(scientific_text):f}"
        
        exp_num = int(exp_num)
        if exp_num == 0:
            return f"{Decimal(scientific_text):f}"
        
        return f"{Decimal(numeric_repr) * (Decimal(10))**exp_num:f}"
    
    def keyPressEvent(self, event: QtGui.QKeyEvent | None) -> None:
        """
        Intercepts the key press event.

        If self._lastCursorPosition is exactly at the end of
        the text, resets self._lastCursorPosition to 0 avoiding
        corrections of cursor position when editing the text.

        Args:
            event (QKeyEvent): the key press event.
        """
        self._lastCursorPosition = self.lineEdit().cursorPosition()
        if self._lastCursorPosition > len(self.lineEdit().text()) - 1:
            # In case we're typing a new number, we shouldn't maintain cursor position.
            self._lastCursorPosition = 0

        return super().keyPressEvent(event)
    
    def keepCursorPosition(self,
                           last_position = None, 
                           correct = False,
                           is_step_by = False,
                           is_editing_finished = False):
        """
        Sets the cursor position to the last position.

        Note:
            - If last_position is not provided, uses the internal last cursor position.
            - If correct is True, corrects the cursor position using correctCursorPosition.
            - The is_step_by and is_editing_finished args are passed to correctCursorPosition.

        Args:
            last_position (int, optional): the last cursor position.
                Defaults to self._lastCursorPosition.
            correct (bool, optional): whether to correct the cursor position.
                Defaults to False.
            is_step_by (bool, optional): whether the cursor position is being
                changed by a step. Passed to correctCursorPosition. Defaults to False.
            is_editing_finished (bool, optional): whether the editing is finished.
                Passed to correctCursorPosition. Defaults to False.
        """
        self.blockSignals(True)
        if last_position is None:
            last_position = self._lastCursorPosition

        if correct:
            self.lineEdit().setCursorPosition(
                self.correctCursorPosition(
                    pre_position = last_position,
                    is_step_by = is_step_by,
                    is_editing_finished = is_editing_finished
                )
            )
        else:
            self.lineEdit().setCursorPosition(last_position)
        self.blockSignals(False)
    
    def textFromValue(self, value: Decimal) -> str:
        """
        Retrieves the text from current value and sets the internal text.

        Side effects:
            - sets the number of decimals
            - corrects the cursor position

        Args:
            value (Decimal): the current numeric value of the widget.
        
        Returns:
            str: the text representation of the value.
        """
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(self._lastText)
        )

        numeric = str(self._value)
        unit = captured.unitText

        if self._enableDimensionless:
            unit = ''

        self.correctDecimals(text=numeric)
        result = ''

        if not self._enableDisplayUnit:
            if not self._enablePreferredUnits:
                result = self.defaultConversor(numeric, unit)
            else:
                result = self.nearestPreferredUnitConversor(numeric, unit)
        else:
            result = self.displayUnitConversor(numeric, unit)

        return self.convertInternalToText(result)
    
    def defaultConversor(self, numeric: str, unit: str) -> str:
        """
        Converts a quantity to its internal text representation.
        
        Args:
            numeric (str): text representation of the numeric part.
            unit (str): text representation of the unit part.
            
        Returns:
            str: the internal text representation.
        """
        # Doesn't do any conversions
        result = f"{Decimal(numeric):.{self._decimals}f}"
        #_logger.debug(f"Converted using defaultConversor. Params: {numeric=}, {unit=}; vars: {self._decimals=}, {result=}")

        if unit:
            result += UNIT_SEPARATOR + unit

        # Sets text, value and corrects cursor position 
        result = self.addSeparators(result)
        self._internalText = result
        self._lastText = self.convertInternalToText(self._internalText)
        self.blockSignals(True)
        self.setValue(Decimal(numeric))
        self.blockSignals(False)
        self.keepCursorPosition(correct=True)
        return self._internalText
    
    def nearestPreferredUnitConversor(self, numeric: str, unit: str) -> str:
        """
        Converts a quantity to its internal text representation with
        the nearest preferred unit.
        
        Args:
            numeric (str): text representation of the numeric part.
            unit (str): text representation of the unit part.
            
        Returns:
            str: the internal text representation in the nearest preferred unit.
        """
        # Converts to nearest preferred unit
        numeric = f"{Decimal(numeric):f}"
        old_quantity = self._backend.quantityFromText(numeric, unit)
        new_quantity = convert_to_nearest_preferred_unit(
            old_quantity,
            self._preferredUnits,
            UNIT_SEPARATOR,
            self._defaultDecimals,
            self._inputValidator,
            self._backend
        )
        new_unit = self._backend.getQuantityUnitStr(new_quantity)
        new_quantity = self._backend.changeQuantityUnit(old_quantity, new_unit, formatter=self.mockedQuantityChangeUnitFormatter)

        #_logger.debug(f"Converted using nearest_preferred: {old_quantity=} -> {new_quantity=}")

        self._quantity = new_quantity
        new_text_repr = self._backend.quantityTextRepr(new_quantity, self._internalUnitSeparator)
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(new_text_repr)
        )

        # Corrects decimals
        numeric = f"{Decimal(captured.scientificRepresentation):f}"
        self.correctDecimals(text=numeric)

        # Sets text, value and corrects cursor position 
        self._internalText = self.addSeparators(new_text_repr)
        self._lastText = self.convertInternalToText(self._internalText)
        self.blockSignals(True)
        self.setValue(Decimal(numeric))
        self.blockSignals(False)
        self.keepCursorPosition(correct=True)
        return self._internalText

    def displayUnitConversor(self, numeric: str, unit: str) -> str:
        """
        Converts a quantity to its internal text representation in the
        display unit.
        
        Args:
            numeric (str): text representation of the numeric part.
            unit (str): text representation of the unit part.
            
        Returns:
            str: the internal text representation in the display unit.
        """
        # Converts quantity to display unit
        new_quantity = self._backend.quantityFromText(numeric, unit)
        new_quantity = self._backend.changeQuantityUnit(new_quantity, self._displayUnit, formatter=self.mockedQuantityChangeUnitFormatter)
        self._quantity = new_quantity
        new_text_repr = self._backend.quantityTextRepr(new_quantity, self._internalUnitSeparator)
    
        # Corrects decimals
        captured = self._inputValidator.getCapturedGroups(
            self.removeSeparators(new_text_repr)
        )
        numeric = f"{Decimal(captured.scientificRepresentation):f}"
        self.correctDecimals(text=numeric)

        # Sets text, value and corrects cursor position
        self._internalText = self.addSeparators(numeric + self._displayUnit)
        self._lastText = self.convertInternalToText(self._internalText)
        self.blockSignals(True)
        self.setValue(Decimal(numeric))
        self.blockSignals(False)
        self.keepCursorPosition(correct=True)
        return self._internalText


"""ScientificSpinBox package.

A Qt5 Widget based on QDoubleSpinBox that enables users to
insert and manipulate physical quantities naturally.

This init file sets up the application.
"""

import os
import logging
import logging.config

# Sets the default interface
from .backend.interfaces import PintInterface
from .defaults import set_default_interface
set_default_interface(PintInterface(unit_system='SI', precision=30))

__version__ = "1.0.0b1"

use_own_logger = os.getenv('SCIENTIFIC_SPINBOX_USE_OWN_LOGGER', 0)

if bool(use_own_logger):
    logging_level = os.getenv('SCIENTIFIC_SPINBOX_LOGGING_LEVEL', 'ERROR')
    if logging_level not in [
        'DEBUG',
        'INFO',
        'ERROR',
        'WARNING',
        'CRITICAL',
    ]:
        raise Exception(f"Invalid logging level `{logging_level}`")

    level_dict = {
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'ERROR': logging.ERROR,
        'WARNING': logging.WARNING,
        'CRITICAL': logging.CRITICAL
    }

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level_dict[logging_level])

    logging.basicConfig(
        level=level_dict[logging_level],
        format='%(asctime)s|%(levelname)s|%(module)s|%(funcName)s|: %(message)s',
        handlers=[logging.FileHandler("test_gui.log", mode='w'), stream_handler]
    )



# PintInterface errors
class IncompatibleConversionUnitError(Exception):
    """
    Exception raised when the unit to be converted to is incompatible
    with the current unit.
    """
    def __init__(self):
        self._message = "The new unit must have the same dimensionality of the current unit."
        super().__init__(self._message)

class EmptyArrayError(Exception):
    """
    Exception raised when an empty array is passed.
    """
    def __init__(self):
        self._message = "The array must have at least one element."
        super().__init__(self._message)

class InvalidUnitError(Exception):
    """
    Exception raised when an invalid unit is passed.
    """
    def __init__(self, unit: str):
        self._message = f"The unit `{unit}` is invalid."
        super().__init__(self._message)




"""
Backend module.

Provides the interfaces between unit manager and ScientificSpinBox
and other resources that shouldn't be coupled with the UI.

Since:
  2024/02/06

Authors:
  - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""


# python
# -*- coding: utf-8 -*-

"""Setup module for ScientificSpinbox.

Since: 2019/02/07

Authors:
    - Eduardo Rocha Falvo <dudu.falvo@gmail.com>
    - Daniel C. Pizetta <daniel.pizetta@usp.br>
    - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""

import codecs
import logging
import os
import re

from setuptools import setup, find_packages

_logger = logging.getLogger(__name__)

def find_version(*file_paths):
    """Find version in a Python file, searching for the __version__."""
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def read(*parts):
    """Read and return the contents of a file."""
    # intentionally *not* adding an encoding option to open, See:
    # https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    here = os.path.abspath(os.path.dirname(__file__))
    return codecs.open(os.path.join(here, *parts), 'r').read()

_version = find_version("scientific_spinbox", "__init__.py")
_long_description = read('README.rst')

_classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Plugins',
    'Intended Audience :: Science/Research',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Natural Language :: English',
    'Operating System :: MacOS',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: User Interfaces',
    'Topic :: Text Editors :: Text Processing',
]

_requirements = [
    "pint",
    "pyqt5",
    "qtpy"
]

_setup_requirements = []

setup(name='scientific-spinbox',
      version=_version,
      description='ScientificSpinbox - A Qt widget to manipulate physical quantities',
      author=['Breno Pelegrin'],
      author_email='brenohqsilva@gmail.com',
      maintainer=['Breno Pelegrin <brenohqsilva@gmail.com>', 'Daniel Cosmo Pizetta <daniel.pizetta@alumni.usp.br>'],
      maintainer_email=['brenohqsilva@gmail.com', 'daniel.pizetta@alumni.usp.br'],
      classifiers=_classifiers,
      long_description=_long_description,
      long_description_content_type='text/x-rst',
      license='LGPLv3',
      license_file='LICENSE.rst',
      packages=find_packages(),
      install_requires=_requirements,
      setup_requires=_setup_requirements,
      project_urls={
          "Homepage": "https://dpizetta.gitlab.io/pqwidget",
          "Documentation": "https://pqwidget.readthedocs.io/",
          "Source": "https://gitlab.com/dpizetta/pqwidget",
          "Issues": "https://gitlab.com/dpizetta/pqwidget/issues"
      }
      )



# ScientificInputValidator errors
class CaptureGroupNameNotFoundOnRegex(Exception):
    """
    Exception raised when capture group name is not found on the
    ScientificInputValidator's regex.
    """
    def __init__(self, group_name: str):
        self._message = f"The capture group `{group_name}` doesn't exist on ScientificInputValidator's regex."
        super().__init__(self._message)

class InvalidValidatorStateError(Exception):
    """
    Exception raised when an invalid validator state is provided.
    """
    def __init__(self, state):
        self._message = f"Invalid state `{state}`."
        super().__init__(self._message)


"""
Validators module.

Provides the text validators and text processing utilities
to be used within a ScientificSpinBox.

Since:
  2024/02/06

Authors:
  - Breno H. Pelegrin S. <breno.pelegrin@usp.br>
"""

from qtpy import QtGui, PYSIDE2
from typing import Type

from .errors.validators import (
    CaptureGroupNameNotFoundOnRegex,
    InvalidValidatorStateError
)
if PYSIDE2:
    from qtpy.QtGui import QRegExpValidator
    from qtpy.QtCore import QRegExp
    QRegularExpression = QRegExp
    QRegularExpressionValidator = QRegExpValidator
    raise NotImplementedError(
        "Compatibility layer for `pyside2` is not yet implemented. This module is only compatible with `pyqt5`, `pyqt6` and `pyside6`."
    )
else:
    from qtpy.QtGui import QRegularExpressionValidator
    from qtpy.QtCore import QRegularExpression

class ScientificInputValidator(QRegularExpressionValidator):
    """Input validator for a ScientificSpinBox.
    
    Verifies if the provided input is a valid scientific number with
    or without units based on a regular expression.

    Implements methods for manipulating captured groups to extract
    parts from the scientific input.
    """
    # Please note that QRegularExpression isn't compatible with Qt4.
    # The new QRegularExpression implements match group names that are
    # useful for this application.
    def __init__(self, parent=None):
        """
        Args:
            parent (object): parent Qt object
        """
        # Allows only patterns like "-0123456789.0123456789E+0123456789 /aBCµΩºde/FG***HI*Jkl*m*/no"

        # (?<full>(?<sci_repr>^(?<num_repr>(?<num_sgn>[+-])?(?<int_num>[0-9]+)?(?<dec_repr>(?<dec_sep>[.,])?(?<dec_num>[0-9]+))?)(?<exp_repr>[eE](?<exp_num>(?<exp_sgn>[+-])[0-9]+))?)(?<txt_repr> *(?<txt_unit>\µ?\Ω?\º?[a-zA-Z0-9?\/*\**\(*\)*]+))?)$
        self._regex = QRegularExpression(
            r"^(?<full>(?<sci_repr>^(?<num_repr>(?<num_sgn>[+-])?(?<int_num>[0-9]+)?(?<dec_repr>(?<dec_sep>[.])?(?<dec_num>[0-9]+))?)(?<exp_repr>[eE](?<exp_num>(?<exp_sgn>[+-])[0-9]+))?)(?<txt_repr> *(?<txt_unit>\µ?\Ω?\º?[a-zA-Z0-9?\/*\**\(*\)*]+))?)$"
        )
        super().__init__(self._regex, parent)

    @property
    def regExpCaptureGroups(self):
        """
        Returns the named capture groups from the regex.

        Returns:
            list: list of named capture groups.
        """
        groups = self._regex.namedCaptureGroups().copy()
        groups.pop(0)
        return groups
        
    def getCapturedGroups(self, text):
        """Returns a ScientificCapturedGroups object for the given text.
        
        Args:
            text (str): text to fill the capture groups.

        Returns:
            ScientificCapturedGroups: object with the captured groups.
        """
        return ScientificCapturedGroups(text, self)

    def stateToStr(self, state: QtGui.QValidator.State):
        """
        Transforms a QValidator state into a string.

        Args:
            state (QtGui.QValidator.State): QValidator state.
            
        Raises:
            InvalidValidatorStateError: if the provided state is not one of the
                known QValidator states (Acceptable, Invalid, Intermediate).
            
        Returns:
            str: string representation of the QValidator state.
                Can be `Acceptable`, `Invalid` or `Intermediate`.
        """
        valid_states = [
            QtGui.QValidator.State.Acceptable,
            QtGui.QValidator.State.Invalid,
            QtGui.QValidator.State.Intermediate
        ]

        if state not in valid_states:
            raise InvalidValidatorStateError(state)

        if state == QtGui.QValidator.State.Intermediate:
            return 'Intermediate'
        elif state == QtGui.QValidator.State.Invalid:
            return 'Invalid'
        elif state == QtGui.QValidator.State.Acceptable:
            return 'Acceptable'

    def getRegExpGroupIdx(self, group_name: str):
        """
        Gets the index of a named capture group based on its name.

        Args:
            group_name (str): name of the group to get index.
        
        Returns:
            int: index of the group.

        Raises:
            Exception: if the group doesn't exist on the regex.
        """
        groups = self.regExpCaptureGroups
        if group_name in groups:
            return groups.index(group_name)
        raise CaptureGroupNameNotFoundOnRegex(group_name)
    
    def interpretText(self, text):
        """
        Replaces commas with dots on the text.

        Args:
            text (str): text to replace commas.

        Returns:
            str: text with only dots.
        """
        text = str(text).replace(',', '.')
        return text
    
    def getAllCaptured(self, text):
        """
        Returns a dictionary with all the capture groups and
        the corresponding text captured by each group.

        Args:
            text (str): text to fill the capture groups.
        """
        captured_dict = {}
        captured_dict = dict.fromkeys(self.regExpCaptureGroups)

        text = self.interpretText(text)
        match = self._regex.match(text)
        for key in captured_dict:
            captured_dict[key] = match.captured(key)

        return captured_dict

    def validate(self, text, position):
        """
        Reimplements the validate method from the parent class.
        
        On the current implementation, it just returns the result from
        the parent class method.

        Args:
            text (str): text to validate.
            position (int): position of the cursor before validation.

        Returns:
            state (QtGui.QValidator.State): state of the validator.
            text (str): text after the validation.
            position (int): position of the cursor after validation.
        """
        pre_state, text, position = super().validate(text, position)

        return (pre_state, text, position)

    def fixup(self, text):
        """
        Reimplements the fixup method from the parent class.
        
        On the current implementation, it just returns the result from
        the parent class method.

        Args:
            text (str): text to fixup.

        Returns:
            str: fixed up text.
        """
        return super().fixup(text)
    
class ScientificCapturedGroups:
    """Class that represents captured groups for a given text.
    
    Implements methods to access each group with ease.
    """
    def __init__(self, text: str, validator: Type[ScientificInputValidator]):
        """
        Args:
            text (str): text to fill the capture groups.
            validator (ScientificInputValidator): validator object reference.
        """
        self._inputText = text
        self._captured = validator.getAllCaptured(text)

    @property
    def fullText(self):
        """Full captured text."""
        return self._captured['full']

    @property
    def scientificRepresentation(self):
        """Scientific representation, without unit text.
        
        Can contain:
            - numeric signal
            - integer numbers
            - decimal separator
            - decimal numbers
            - exponential symbol
            - exponential signal
            - exponential numbers.

        Doesn't contain:
            - unit text
        """
        return self._captured['sci_repr']
    
    # <num_repr>
    @property
    def numericRepresentation(self):
        """Numeric representation, without exponential.
        
        Can contain:
            - numeric signal
            - integer numbers
            - decimal separator
            - decimal numbers

        Doesn't contain:
            - exponential symbol
            - exponential signal
            - exponential numbers
            - unit text
        """
        return self._captured['num_repr']

    @property
    def numericSignal(self):
        """Signal of the numeric part."""
        return self._captured['num_sgn']

    @property
    def integerNumbers(self):
        """Integer part, not including signal."""
        return self._captured['int_num']
    
    # <dec_repr>
    @property
    def decimalRepresentation(self):
        """Decimal part, including separator and numbers."""
        return self._captured['dec_repr']
    
    @property
    def decimalSeparator(self):
        """Decimal separator."""
        return self._captured['dec_sep']
    
    @property
    def decimalNumbers(self):
        """Numbers of the decimal part."""
        return self._captured['dec_num']
    
    @property
    def exponentialRepresentation(self):
        """Exponential part, including exponential sign, number and signal."""
        return self._captured['exp_repr']
    
    @property
    def exponentialNumber(self):
        """Number after the exponential sign, including signal."""
        return self._captured['exp_num']
    
    @property
    def exponentialSignal(self):
        """Signal of the exponential."""
        return self._captured['exp_sgn']
    
    @property
    def unitRepresentation(self):
        """Unit part, with the preceding spaces."""
        return self._captured['txt_repr']

    @property
    def unitText(self):
        """Text of the unit part, without the preceding spaces."""
        return self._captured['txt_unit']


from qtpy import QtWidgets, QtCore

from scientific_spinbox.widget import ScientificSpinBox
from test_gui.ui.dlg_testspinbox_shortexample_gui import Ui_Dialog as Ui_ShortExampleGUI

class ShortExampleGUIDialog(QtWidgets.QDialog):
    """
    Dialog that shows ScientificSpinBox and some test controls.

    Used to test the behaviour of the ScientificSpinBox GUI.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui = Ui_ShortExampleGUI()

        # Configures Ui
        self.ui.setupUi(self)

        self.scientificSpinBox = ScientificSpinBox(
            base_unit='ms',
            allowed_units=['ms', 's', 'us'],
            parent=self.ui.testSpinBox1.parentWidget()
        )

        # Replaces necessary widgets
        self.ui.testSpinBox1.parentWidget().layout().replaceWidget(
            self.ui.testSpinBox1,
            self.scientificSpinBox
        )
        self.ui.testSpinBox1 = self.scientificSpinBox
        self.scientificSpinBox.valueChanged.connect(self.onValueChangeSlot)
        self.scientificSpinBox.baseValueChanged.connect(self.onBaseValueChangeSlot)

    def onValueChangeSlot(self):
        self.ui.valueLabel.setText(f"value: {str(self.scientificSpinBox.value)}")
        self.ui.unitLabel.setText(f"unit: {str(self.scientificSpinBox.unit)}")
        self.ui.baseUnitLabel.setText(f"baseUnit: {str(self.scientificSpinBox.baseUnit)}")

    def onBaseValueChangeSlot(self):
        self.ui.baseValueLabel.setText(f"baseValue: {str(self.scientificSpinBox.baseValue)}")

    